"""Define Parameter which holds the value."""

#  Copyright (c) 2024 zfit

from __future__ import annotations

import abc
import collections
import copy
import functools
import typing
import warnings
import weakref
from collections.abc import Callable, Iterable
from contextlib import suppress
from inspect import signature
from typing import Literal, Mapping, Optional, Union
from weakref import WeakSet

import dill
import numpy as np
import pydantic
import tensorflow as tf
import tensorflow_probability as tfp

# TF backwards compatibility
from ordered_set import OrderedSet
from pydantic import Field, validator
from tensorflow.python.ops import tensor_getitem_override
from tensorflow.python.ops.resource_variable_ops import (
    ResourceVariable as TFVariable,
)
from tensorflow.python.ops.resource_variable_ops import (
    VariableSpec,
)
from tensorflow.python.ops.variables import Variable
from tensorflow.python.types.core import Tensor as TensorType

import zfit.z.numpy as znp

from .. import z
from ..core.baseobject import BaseNumeric, extract_filter_params
from ..minimizers.interface import ZfitResult
from ..serialization.paramrepr import make_param_constructor
from ..serialization.serializer import BaseRepr, Serializer
from ..settings import run, ztypes
from ..util import ztyping
from ..util.checks import NotSpecified
from ..util.container import convert_to_container
from ..util.deprecation import deprecated, deprecated_args
from ..util.exception import (
    BreakingAPIChangeError,
    FunctionNotImplemented,
    IllegalInGraphModeError,
    LogicalUndefinedOperationError,
    ParameterNotIndependentError,
)
from ..util.temporary import TemporarilySet
from . import interfaces as zinterfaces
from .dependents import _extract_dependencies
from .interfaces import ZfitIndependentParameter, ZfitModel, ZfitParameter
from .serialmixin import SerializableMixin

# todo add type hints in this module for api


class MetaBaseParameter(type(tf.Variable), type(zinterfaces.ZfitParameter)):  # resolve metaclasses
    pass


def register_tensor_conversion(
    convertable, name=None, overload_operators=True, priority=10
):  # higher than any tf conversion
    def _dense_var_to_tensor(var, dtype=None, name=None, as_ref=False):
        return var._dense_var_to_tensor(dtype=dtype, name=name, as_ref=as_ref)

    tf.register_tensor_conversion_function(convertable, _dense_var_to_tensor, priority=priority)
    if name:
        pass
        # _pywrap_utils.RegisterType(name, convertable)

    if overload_operators:
        convertable._OverloadAllOperators()


class OverloadableMixin(ZfitParameter):
    # Conversion to tensor.
    @staticmethod
    def _TensorConversionFunction(v, dtype=None, name=None, as_ref=False):  # pylint: disable=invalid-name
        """Utility function for converting a Variable to a Tensor."""
        _ = name
        if dtype and not dtype.is_compatible_with(v.dtype):
            msg = (
                f"Incompatible type conversion requested to type '{dtype.name}' for variable "
                f"of type '{v.dtype.name}'"
            )
            raise ValueError(msg)
        if as_ref:
            return v._ref()  # pylint: disable=protected-access
        else:
            return v.value()

    def _dense_var_to_tensor(self, dtype=None, name=None, as_ref=False):
        del name
        if dtype and not dtype.is_compatible_with(self.dtype):
            msg = (
                f"Incompatible type conversion requested to type '{dtype.name}' for variable "
                f"of type '{self.dtype.name}'"
            )
            raise ValueError(msg)
        if as_ref:
            if hasattr(self, "_ref"):
                return self._ref()
            else:
                msg = "Why is this needed?"
                raise RuntimeError(msg)
        else:
            return self.value()

    def _AsTensor(self):
        return self.value()

    @classmethod
    def _OverloadAllOperators(cls):  # pylint: disable=invalid-name
        """Register overloads for all operators."""
        for operator in tf.Tensor.OVERLOADABLE_OPERATORS:
            cls._OverloadOperator(operator)
        # For slicing, bind getitem differently than a tensor (use SliceHelperVar
        # instead)
        # pylint: disable=protected-access
        cls.__getitem__ = tensor_getitem_override._slice_helper_var

    @classmethod
    def _OverloadOperator(cls, operator):  # pylint: disable=invalid-name
        """Defer an operator overload to ``ops.Tensor``.

        We pull the operator out of ops.Tensor dynamically to avoid ordering issues.
        Args:
          operator: string. The operator name.
        """
        # We can't use the overload mechanism on __eq__ & __ne__ since __eq__ is
        # called when adding a variable to sets. As a result we call a.value() which
        # causes infinite recursion when operating within a GradientTape
        # TODO(gjn): Consider removing this
        if operator in ("__eq__", "__ne__"):
            return

        tensor_oper = getattr(tf.Tensor, operator)

        def _run_op(a, *args, **kwargs):
            # pylint: disable=protected-access
            return tensor_oper(a.value(), *args, **kwargs)

        functools.update_wrapper(_run_op, tensor_oper)
        setattr(cls, operator, _run_op)


register_tensor_conversion(OverloadableMixin, overload_operators=True)


class WrappedVariable(metaclass=MetaBaseParameter):
    def __init__(self, initial_value, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.variable = tf.Variable(
            initial_value=initial_value,
            name=self.name,
            dtype=self.dtype,
        )

    @property
    @abc.abstractmethod
    def name(self):
        raise NotImplementedError

    @property
    def constraint(self):
        return self.variable.constraint

    @property
    def dtype(self):
        return self.variable.dtype

    def value(self):
        return self.variable.value()

    def read_value(self):
        return self.variable.read_value()

    @property
    def shape(self):
        return self.variable.shape

    def numpy(self):
        return self.variable.numpy()

    def assign(self, value, use_locking=False, name=None, read_value=True):
        return self.variable.assign(value=value, use_locking=use_locking, name=name, read_value=read_value)

    def _dense_var_to_tensor(self, dtype=None, name=None, as_ref=False):
        del name
        if dtype is not None and dtype != self.dtype:
            return NotImplemented
        if as_ref:
            return self.variable.read_value().op.inputs[0]
        else:
            return self.variable.value()

    def _AsTensor(self):
        return self.variable.value()

    @staticmethod
    def _OverloadAllOperators():  # pylint: disable=invalid-name
        """Register overloads for all operators."""
        for operator in tf.Tensor.OVERLOADABLE_OPERATORS:
            WrappedVariable._OverloadOperator(operator)
        # For slicing, bind getitem differently than a tensor (use SliceHelperVar
        # instead)
        # pylint: disable=protected-access
        WrappedVariable.__getitem__ = tensor_getitem_override._slice_helper_var

    @staticmethod
    def _OverloadOperator(operator):  # pylint: disable=invalid-name
        """Defer an operator overload to ``ops.Tensor``.

        We pull the operator out of ops.Tensor dynamically to avoid ordering issues.
        Args:
          operator: string. The operator name.
        """
        tensor_oper = getattr(tf.Tensor, operator)

        def _run_op(a, *args):
            # pylint: disable=protected-access
            value = a._AsTensor()
            return tensor_oper(value, *args)

        # Propagate __doc__ to wrapper
        with suppress(AttributeError):
            _run_op.__doc__ = tensor_oper.__doc__

        setattr(WrappedVariable, operator, _run_op)


register_tensor_conversion(WrappedVariable, "WrappedVariable", overload_operators=True)


class BaseParameter(Variable, ZfitParameter, TensorType, metaclass=MetaBaseParameter):
    def __init__(self, *args, **kwargs):
        try:
            super().__init__(*args, **kwargs)
        except NotImplementedError as error:
            tmp_val = kwargs.pop("name", None)  # remove if name is in there, needs to be passed through
            if args or kwargs:
                kwargs["name"] = tmp_val
                msg = (
                    f"The following arguments reached the top of the inheritance tree, the super "
                    f"init is not implemented (most probably abstract tf.Variable): {args, kwargs}. "
                    f"If you see this error, please post it as an bug at: "
                    f"https://github.com/zfit/zfit/issues/new/choose"
                )
                raise RuntimeError(msg) from error

    def __len__(self):
        return 1


class ZfitParameterMixin(BaseNumeric):
    _existing_params: typing.ClassVar = {}

    def __init__(self, name, label=None, **kwargs):
        if name not in self._existing_params:
            self._existing_params[name] = WeakSet()
            # Is an alternative arg for pop needed in case it fails? Why would it fail?
            weakref.finalize(
                self,
                lambda name: name in self._existing_params and self._existing_params.pop,
                name,
            )
        self._existing_params[name].add(self)
        self._name = name
        self._label = label

        super().__init__(name=name, **kwargs)
        self._assert_params_unique()

    # property needed here to overwrite the name of tf.Variable
    @property
    def name(self) -> str:
        return self._name

    @property
    def label(self) -> str:
        if (label := self._label) is None:
            label = self.name
        return label

    def __del__(self):
        with suppress(AttributeError, NotImplementedError):  # if super does not have a __del__
            super().__del__(self)

    def __add__(self, other):
        if isinstance(other, (ZfitModel, ZfitParameter)):
            from . import operations

            with suppress(FunctionNotImplemented):
                return operations.add(self, other)
        return super().__add__(other)

    def __radd__(self, other):
        if isinstance(other, (ZfitModel, ZfitParameter)):
            from . import operations

            with suppress(FunctionNotImplemented):
                return operations.add(other, self)
        return super().__radd__(other)

    def __mul__(self, other):
        if isinstance(other, (ZfitModel, ZfitParameter)):
            from . import operations

            with suppress(FunctionNotImplemented):
                return operations.multiply(self, other)
        return super().__mul__(other)

    def __rmul__(self, other):
        if isinstance(other, (ZfitModel, ZfitParameter)):
            from . import operations

            with suppress(FunctionNotImplemented):
                return operations.multiply(other, self)
        return super().__rmul__(other)

    def __eq__(self, other):
        return self is other

    def __hash__(self):
        if not hasattr(self, "_cached_hash"):
            self._cached_hash = hash(id(self))
        return self._cached_hash


class TFBaseVariable(TFVariable, metaclass=MetaBaseParameter):
    # class TFBaseVariable(WrappedVariable, metaclass=MetaBaseParameter):

    # Needed, otherwise tf variable complains about the name not having a ':' in there
    @property
    def _shared_name(self):
        return self.name


class Parameter(
    ZfitParameterMixin,
    TFBaseVariable,
    BaseParameter,
    SerializableMixin,
    ZfitIndependentParameter,
):
    """Class for fit parameters that has a default state."""

    _independent = True
    _independent_params = WeakSet()
    DEFAULT_STEP_SIZE = 0.01

    def __init__(
        self,
        name: str,
        value: ztyping.NumericalScalarType,
        lower: ztyping.NumericalScalarType | None = None,
        upper: ztyping.NumericalScalarType | None = None,
        step_size: ztyping.NumericalScalarType | None = None,
        floating: bool = True,
        *,
        dtype: tf.DType = None,
        label: str | None = None,
        # legacy
        lower_limit: ztyping.NumericalScalarType | None = None,
        upper_limit: ztyping.NumericalScalarType | None = None,
    ):
        """Fit Parameter that has a default state (value) and limits (lower, upper).

        The name identifies the parameter.
        Multiple parameters with the same name can exist, however,
        they cannot be in the same PDF/func/loss as the value would not be uniquely defined.

        Args:
            name : Name of the parameter. Should be unique within a model/likelihood.
            value : Default value of the parameter. Also used as the starting value in minimization.
            lower : lower limit of the parameter. If the parameter is set to a value below the lower limit, it will raise an error.
            upper : upper limit of the parameter. If the parameter is set to a value above the upper limit, it will raise an error.
            floating : If the parameter is floating (can change value) or fixed (constant) in the minimization.
            label: |@doc:param.init.label||@docend:param.init.label|
            step_size : Initial step size for minimization. If not set, a default value is used.
        """
        self._independent_params.add(self)

        # legacy start
        if lower_limit is not None:
            msg = "The argument `lower_limit` has been renamed to `lower`."
            raise BreakingAPIChangeError(msg)
        if upper_limit is not None:
            msg = "The argument `upper_limit` has been renamed to `upper`."
            raise BreakingAPIChangeError(msg)
        if dtype is None:
            dtype = ztypes.float
        else:
            warnings.warn(
                "The argument `dtype` is deprecated and will be removed in the future.",
                DeprecationWarning,
                stacklevel=2,
            )
        # legacy end

        # TODO: sanitize input for TF2
        self._lower_limit_neg_inf = None
        self._upper_limit_neg_inf = None
        if lower is None:
            self._lower_limit_neg_inf = znp.asarray(-np.inf, dtype)
        if upper is None:
            self._upper_limit_neg_inf = znp.asarray(np.inf, dtype)
        value = znp.asarray(value, dtype=ztypes.float)

        super().__init__(
            initial_value=value,
            dtype=dtype,
            name=name,
            params={},
            label=label,
        )

        self.lower = lower
        self.upper = upper
        self.floating = floating
        self.step_size = step_size
        self.set_value(value)  # to check that it is in the limits

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._independent = True  # overwriting independent only for subclass/instance

    @classmethod
    def _from_name(cls, name):
        for param in cls._independent_params:
            if name == param.name:
                return param
        msg = f"Parameter {name} does not exist, please create it first."
        raise ValueError(msg)

    @property
    def lower(self):
        limit = self._lower
        if limit is None:
            limit = self._lower_limit_neg_inf
        return limit

    @lower.setter
    # @invalidate_graph
    def lower(self, value):
        if value is None and self._lower_limit_neg_inf is None:
            self._lower_limit_neg_inf = znp.asarray(-np.inf, dtype=ztypes.float)
        elif value is not None:
            value = znp.asarray(value, dtype=ztypes.float)
        self._lower = value

    @property
    def upper(self):
        limit = self._upper
        if limit is None:
            limit = self._upper_limit_neg_inf
        return limit

    @upper.setter
    # @invalidate_graph
    def upper(self, value):
        if value is None and self._upper_limit_neg_inf is None:
            self._upper_limit_neg_inf = znp.asarray(np.inf, dtype=ztypes.float)
        elif value is not None:
            value = znp.asarray(value, dtype=ztypes.float)
        self._upper = value

    @property
    def has_limits(self) -> bool:
        """If the parameter has limits set or not."""
        no_limits = self._lower is None and self._upper is None
        return not no_limits

    @property
    def at_limit(self) -> tf.Tensor:
        """If the value is at the limit (or over it).

        Returns:
            Boolean ``tf.Tensor`` that tells whether the value is at the limits.
        """
        return self._check_at_limit(self.value())

    def _check_at_limit(self, value, exact=False):
        """The precision is up to 1e-5 relative or 1e-8 absolute if exact is None.

        Args:
            value ():
            exact ():

        Returns:
        """
        if not self.has_limits:
            return tf.constant(False)
        # Adding a slight tolerance to make sure we're not tricked by numerics due to floating point comparison
        diff = znp.abs(self.upper - self.lower)  # catch if it is minus inf
        if not exact:
            reltol = 0.005
            abstol = 1e-5
        else:
            reltol = 1e-5
            abstol = 1e-7
        tol = znp.minimum(diff * reltol, abstol)  # if one limit is inf we would get inf
        if not exact:  # if exact, we wanna allow to set it slightly over the limit.
            tol = -tol  # If not, we wanna make sure it's inside
        at_lower = z.unstable.less_equal(value, self.lower - tol)
        at_upper = z.unstable.greater_equal(value, self.upper + tol)
        return z.unstable.logical_or(at_lower, at_upper)

    def value(self):
        value = super().value()
        if self.has_limits:
            value = tfp.math.clip_by_value_preserve_gradient(
                value, clip_value_min=self.lower, clip_value_max=self.upper
            )
        return value

    @deprecated(None, "Use `value` instead.")
    def read_value(self):
        return self.value()

    @property
    def floating(self):
        if self._floating and (hasattr(self, "trainable") and not self.trainable):
            msg = "Floating is set to true but tf Variable is not trainable."
            raise RuntimeError(msg)
        return self._floating

    @floating.setter
    def floating(self, value):
        if not isinstance(value, bool):
            msg = "floating has to be a boolean."
            raise TypeError(msg)
        self._floating = value

    def _get_dependencies(self):
        return {self}

    @property
    def independent(self):
        return self._independent

    @property
    def has_step_size(self):
        return self._step_size is not None

    @property
    def step_size(self) -> tf.Tensor:  # TODO: improve default step_size?
        """Step size of the parameter, the estimated order of magnitude of the uncertainty.

        This can be crucial to tune for the minimization. A too large ``step_size`` can produce NaNs, a too small won't
        converge.

        If the step size is not set, the ``DEFAULT_STEP_SIZE`` is used.

        Returns:
            The step size
        """
        step_size = self._step_size
        if step_size is None:
            #     # auto-infer from limits
            #     step_splits = 1e5
            #     if self.has_limits:
            #         step_size = (self.upper_limit - self.lower_limit) / step_splits  # TODO improve? can be tensor?
            #     else:
            #         step_size = self.DEFAULT_STEP_SIZE
            #     if np.isnan(step_size):
            #         if self.lower_limit == -np.infty or self.upper_limit == np.infty:
            #             step_size = self.DEFAULT_STEP_SIZE
            #         else:
            #             raise ValueError("Could not set step size. Is NaN.")
            #     # step_size = z.to_real(step_size)
            #     self.step_size = step_size
            step_size = self.DEFAULT_STEP_SIZE
        # step_size = z.convert_to_tensor(step_size)
        return step_size

    @step_size.setter
    def step_size(self, value):
        if value is not None:
            value = float(value)
            # value = z.convert_to_tensor(value, preferred_dtype=ztypes.float)
            # value = znp.asarray(value, dtype=ztypes.float)
        self._step_size = value

    def set_value(self, value: ztyping.NumericalScalarType):
        """Set the :py:class:`~zfit.Parameter` to `value` (temporarily if used in a context manager).

        This operation won't, compared to the assign, return the read value but an object that *can* act as a context
        manager.

        Args:
            value: The value the parameter will take on.
        Raises:
            ValueError: If the value is not inside the limits (in normal Python/eager mode)
            InvalidArgumentError: If the value is not inside the limits (in JIT/traced/graph mode)
        """

        def getter():
            return self.value()

        def setter(value):
            if self.has_limits:
                message = (
                    f"Setting value {value} invalid for parameter {self.name} with limits "
                    f"{self.lower} - {self.upper}. This is changed."
                    f" In order to silence this and clip the value, you can use (with caution,"
                    f" advanced) `Parameter.assign`"
                )
                if run.executing_eagerly():
                    if self._check_at_limit(value, exact=True):
                        raise ValueError(message)
                else:
                    tf.debugging.assert_greater(
                        znp.asarray(value, tf.float64),
                        znp.asarray(self.lower, tf.float64),
                        message=message,
                    )
                    tf.debugging.assert_less(
                        znp.asarray(value, tf.float64),
                        znp.asarray(self.upper, tf.float64),
                        message=message,
                    )
            #     tf.debugging.Assert(self._check_at_limit(value), [value])
            self.assign(value=value, read_value=False)

        return TemporarilySet(value=value, setter=setter, getter=getter)

    def assign(self, value, use_locking: bool | None = None, read_value=False):
        """Set the :py:class:`~zfit.Parameter` to `value` without any checks.

        Compared to ``set_value``, this method cannot be used with a context manager and won't raise an
        error

        Args:
            value: The value the parameter will take on.
        """
        return super().assign(value=value, use_locking=use_locking, read_value=read_value)

    def randomize(
        self,
        minval: ztyping.NumericalScalarType | None = None,
        maxval: ztyping.NumericalScalarType | None = None,
        sampler: Callable = np.random.uniform,
    ) -> tf.Tensor:
        """Update the parameter with a randomised value between minval and maxval and return it.

        Args:
            minval: The lower bound of the sampler. If not given, ``lower_limit`` is used.
            maxval: The upper bound of the sampler. If not given, ``upper_limit`` is used.
            sampler: A sampler with the same interface as ``np.random.uniform``

        Returns:
            The sampled value
        """
        if not tf.executing_eagerly():
            msg = "Randomizing values in a parameter within Graph mode is most probably not" " what is "
            raise IllegalInGraphModeError(msg)
        minval = self.lower if minval is None else znp.asarray(minval, dtype=self.dtype)
        maxval = self.upper if maxval is None else znp.asarray(maxval, dtype=self.dtype)
        if maxval is None or minval is None:
            msg = "Cannot randomize a parameter without limits or limits given."
            raise RuntimeError(msg)
        value = sampler(size=self.shape, low=minval, high=maxval)

        self.set_value(value=value)
        return value

    def get_params(
        self,
        floating: bool | None = True,
        is_yield: bool | None = None,
        extract_independent: bool | None = True,
        only_floating=NotSpecified,
    ) -> set[ZfitParameter]:
        del is_yield, only_floating, extract_independent  # doesnot make sense for a single parameter
        return extract_filter_params(self, floating=floating, extract_independent=False)

    def __repr__(self):  # many try and except in case it's not fully initialized yet
        if tf.executing_eagerly():  # more explicit: we check for exactly this attribute, nothing inside numpy
            try:
                value = f"{self.numpy():.4g}"
            except Exception as err:
                value = f"errored {err}"
        else:
            value = "graph-node"
        try:
            floating = self.floating
        except Exception as err:
            floating = f"errored {err}"
        try:
            name = self.name
        except Exception as err:
            name = f"errored {err}"
        return f"<zfit.{self.__class__.__name__} '{name}' floating={floating} value={value}>"

    # LEGACY, deprecate?

    @property
    @deprecated(None, "Use `lower` instead.")
    def lower_limit(self):
        return self.lower

    @lower_limit.setter
    @deprecated(None, "Use `lower` instead.")
    def lower_limit(self, value):
        self.lower = value

    @property
    @deprecated(None, "Use `upper` instead.")
    def upper_limit(self):
        return self.upper

    @upper_limit.setter
    @deprecated(None, "Use `upper` instead.")
    def upper_limit(self, value):
        self.upper = value

    def __tf_tracing_type__(self, signature_context):
        return ParameterSpec(parameter=self)

    # below needed?
    # CompositeTensor method
    # @property
    # def _type_spec(self):
    #     return ParameterSpec(parameter=self)
    #
    # # CompositeTensor method
    # def _shape_invariant_to_type_spec(self, shape):
    #     return ParameterSpec(shape, self.dtype, self.trainable)

    def __reduce__(self):  # for pickling
        return functools.partial(
            Parameter,
            name=self.name,
            value=self.value(),
            lower=self.lower,
            upper=self.upper,
            floating=self.floating,
            label=self.label,
            step_size=self.step_size,
        ), ()


# delattr(Parameter, "__tf_tracing_type__")


class ParameterSpec(VariableSpec):
    value_type = property(lambda _: Parameter)

    def __init__(self, shape=None, dtype=None, trainable=True, alias_id=None, *, parameter=None):
        if parameter is None:
            msg = "Unknown error, please report, parameter should not be none"
            raise RuntimeError(msg)
        if parameter is not None:  # initialize from parameter
            shape = parameter.shape
            dtype = parameter.dtype
            trainable = True
            alias_id = None
        self.parameter_value = parameter
        self._name = parameter.name
        self.parameter_type = type(parameter)
        if dtype is None:
            dtype = tf.float64
        super().__init__(shape=shape, dtype=dtype, trainable=trainable, alias_id=alias_id)
        self.hash = hash(id(parameter))

    @classmethod
    def from_value(cls, value):
        return cls(parameter=value)

    def _to_components(self, value):
        return super()._to_components(value)

    def _from_components(self, components):
        _ = super()._from_components(components)  # checking that there is no error
        return Parameter._from_name(self.name)

    def _to_tensors(self, value):
        return [value]

    def is_subtype_of(self, other):
        return (
            type(other) is ParameterSpec
            and self.parameter_type is other.parameter_type
            and self.parameter_value is other.parameter_value
        )

    def most_specific_common_supertype(self, others):
        return self if all(self == other for other in others) else None

    def placeholder_value(self, placeholder_context=None):
        del placeholder_context  # unused
        return self.parameter_value

    def is_compatible_with(self, spec_or_value):
        return isinstance(spec_or_value, ParameterSpec) and self.parameter_value is spec_or_value.parameter_value

    def __eq__(self, other) -> bool:
        if not isinstance(other, ParameterSpec):
            return False
        return self.parameter_value is other.parameter_value

    def __hash__(self):
        return self.hash


class ParameterRepr(BaseRepr):  # add label?
    _implementation = Parameter
    _constructor = pydantic.PrivateAttr(make_param_constructor(Parameter))
    hs3_type: Literal["Parameter"] = Field("Parameter", alias="type")
    name: str
    value: float
    lower: Optional[float] = Field(None, alias="min")
    upper: typing.Optional[float] = Field(None, alias="max")
    step_size: Optional[float] = None
    floating: Optional[bool] = None

    @validator("value", pre=True)
    def _validate_value(cls, v):
        if cls.orm_mode(v):
            v = v()
        return v

    def __hash__(self):
        return hash(self.name) + hash(type(self))


class BaseComposedParameter(ZfitParameterMixin, OverloadableMixin, BaseParameter):
    def __init__(self, params, func, dtype=None, name="BaseComposedParameter", **kwargs):
        # 0.4 breaking
        if "value" in kwargs:
            msg = "'value' cannot be provided any longer, `func` is needed."
            raise BreakingAPIChangeError(msg)
        super().__init__(name=name, params=params, **kwargs)
        if not callable(func):
            msg = "`func` is not callable."
            raise TypeError(msg)

        self._func = func
        self._dtype = dtype if dtype is not None else ztypes.float

    def _get_dependencies(self):
        return _extract_dependencies(list(self.params.values()))

    @property
    def floating(self):
        msg = "Cannot be floating or not. Look at the dependencies."
        raise LogicalUndefinedOperationError(msg)

    @floating.setter
    def floating(self, value):  # noqa: ARG002
        msg = "Cannot set floating or not. Set in the dependencies (`get_params`)."
        raise LogicalUndefinedOperationError(msg)

    @property
    def params(self):
        return self._params

    def value(self):
        params = self.params
        return znp.asarray(self._func(params), dtype=self.dtype)

        # return tf.convert_to_tensor(value, dtype=self.dtype)

    def read_value(self):
        return tf.identity(self.value())

    @property
    def shape(self):
        return self.value().shape

    def numpy(self):
        return self.value().numpy()

    @property
    def independent(self):
        return False

    def set_value(self, value):  # noqa: ARG002
        """Set the value of the parameter. Cannot be used for composed parameters!

        Args:
            value:
        """
        msg = "Cannot set value of a composed parameter." " Set the value on its components."
        raise LogicalUndefinedOperationError(msg)

    def randomize(self, minval=None, maxval=None, sampler=np.random.uniform):  # noqa: ARG002
        """Randomize the value of the parameter.

        Cannot be used for composed parameters!
        """
        msg = "Cannot randomize a composed parameter."
        raise LogicalUndefinedOperationError(msg)

    def assign(self, value, use_locking=False, name=None, read_value=True):  # noqa: ARG002
        """Assign the value of the parameter.

        Cannot be used for composed parameters!
        """
        msg = "Cannot assign a composed parameter."
        raise LogicalUndefinedOperationError(msg)


class ConstantParameter(OverloadableMixin, ZfitParameterMixin, BaseParameter, SerializableMixin):
    """Constant parameter.

    Value cannot change.
    """

    def __init__(
        self,
        name,
        value,
        *,
        label: str | None = None,
    ):
        """Constant parameter that cannot change its value.

        Args:
            name: Unique identifier of the parameter.
            value: Constant value.
            label: |@doc:param.init.label||@docend:param.init.label|
        """
        super().__init__(name=name, params={}, dtype=ztypes.float, label=label)
        self._value_np = tf.get_static_value(value, partial=True)
        self._value = tf.guarantee_const(tf.convert_to_tensor(value, dtype=self.dtype))

    @property
    def shape(self):
        return self.value().shape

    def value(self) -> tf.Tensor:
        return self._value

    def read_value(self) -> tf.Tensor:
        return self.value()

    @property
    def floating(self):
        return False

    @floating.setter
    def floating(self, value):  # noqa: ARG002
        msg = "Cannot set a ConstantParameter to floating. Use a `Parameter` instead."
        raise LogicalUndefinedOperationError(msg)

    @property
    def independent(self) -> bool:
        return False

    def _get_dependencies(self) -> ztyping.DependentsType:
        return OrderedSet()

    @property
    def static_value(self):
        return self._value_np

    def numpy(self):
        return self._value_np

    def __repr__(self):
        value_str = f"{value: .4g}" if (value := self._value_np) is not None else "symbolic"
        return f"<zfit.param.{self.__class__.__name__} '{self.name}' dtype={self.dtype.name} value={value_str}>"


register_tensor_conversion(ConstantParameter, "ConstantParameter", overload_operators=True)
register_tensor_conversion(BaseComposedParameter, "BaseComposedParameter", overload_operators=True)


class ConstantParamRepr(BaseRepr):
    _implementation = ConstantParameter
    _constructor = pydantic.PrivateAttr(make_param_constructor(ConstantParameter))
    hs3_type: Literal["ConstantParameter"] = pydantic.Field("ConstantParameter", alias="type")
    name: str
    value: float
    floating: bool = False

    # lower: Optional[float] = Field(None, alias="min")
    # upper: Optional[float] = Field(None, alias="max")

    @validator("value", pre=True)
    def _validate_value(cls, value):
        if cls.orm_mode(value):
            value = value()
        return value

    def _to_orm(self, init):
        init = copy.copy(init)
        init.pop("floating")
        return super()._to_orm(init)


class ComposedParameter(SerializableMixin, BaseComposedParameter):
    @deprecated_args(None, "Use `params` instead.", "dependents")
    @deprecated_args(None, "Use `func` instead.", "value_fn")
    def __init__(
        self,
        name: str,
        value_fn: Optional[Callable] = None,
        *,
        func: Optional[Callable] = None,
        params: (dict[str, ZfitParameter] | Iterable[ZfitParameter] | ZfitParameter) = NotSpecified,
        dtype: tf.dtypes.DType = ztypes.float,
        label: str | None = None,
        unpack_params: bool | None = None,
        dependents: (dict[str, ZfitParameter] | Iterable[ZfitParameter] | ZfitParameter) = NotSpecified,
    ):
        """Arbitrary composition of parameters.

        A `ComposedParameter` allows for arbitrary combinations of parameters and correlations using an arbitrary
        function.

        Examples:
            .. jupyter-execute::

                import zfit

                param1 = zfit.Parameter('param1', 1.0, 0.1, 1.8)
                param2 = zfit.Parameter('param2', 42.0, 0, 100)

                # using a dict for the params
                def mult_dict(params):
                    return params["a"] * params["b"]

                mult_param_dict = zfit.ComposedParameter('mult_dict', mult_dict, params={"a": param1, "b": param2})

                # using a list for the params
                def mult_list(params):
                    return params[0] * params[1]

                mult_param_list = zfit.ComposedParameter('mult_list', mult_list, params=[param1, param2])



        Args:
            name: Unique name of the Parameter.
            func: Function that returns the value of the composed parameter and takes as arguments `params` as
                arguments. The function must be able to be called with the same arguments as `params`.
            params: If it is a `dict`, this will directly be used as the `params` attribute, otherwise the
                parameters will be automatically named with f"param_{i}". The values act as arguments to `func`.
            dtype: Output of `func` dtype
            label: |@doc:param.init.label||@docend:param.init.label|
            unpack_params: If True, the parameters will be unpacked and passed as arguments to `func`. If False, the
                parameters will be passed as a dict/tuple. If None, it will be automatically determined and raise an error
                if it cannot be determined.
            dependents:
                .. deprecated:: unknown
                    use `params` instead.
        """
        # legacy
        if value_fn is not None:
            if func is not None:
                msg = "Cannot specify both `value_fn` and `func`."
                raise ValueError(msg)
            func = value_fn
            del value_fn
        # end legacy
        if not isinstance(params, Mapping):
            params = convert_to_container(params)
        original_init = {"name": name, "internal_params": params, "func": func, "unpack_params": unpack_params}

        # legacy
        if dependents is not NotSpecified:
            params = dependents
        elif params is NotSpecified:
            raise ValueError
        # end legacy

        if params is None:
            msg = "Params needs to be specified."
            raise BreakingAPIChangeError(msg)
        takes_no_params = False
        if unpack_params is None:
            parameters = signature(func).parameters
            if isinstance(params, ZfitParameter):
                params = [params]
                unpack_params = True
            elif len(parameters) == 1 and (len(params) > 1) or "params" in parameters:
                unpack_params = False
            elif len(parameters) - len(params) >= 0:
                unpack_params = True
            elif len(parameters) == 0:
                takes_no_params = True
                unpack_params = False
            else:
                msg = (
                    "Cannot determine if parameter should be unpacked or not. Please specify explicitly `unpack_params`"
                )
                raise ValueError(msg)

        dictlike = isinstance(params, collections.abc.Mapping)

        if dictlike:
            params_dict = params
            if takes_no_params:

                def stratified_fn(params):
                    del params
                    return func()

            elif unpack_params:

                def stratified_fn(params):
                    return func(**params)

            else:
                stratified_fn = func
        else:
            params = convert_to_container(params)

            if params is None or takes_no_params:
                params_dict = {}

                def stratified_fn(params):
                    del params
                    return func()

            else:
                params_dict = {f"param_{i}": p for i, p in enumerate(params)}

                if unpack_params:

                    def stratified_fn(params):
                        return func(*tuple(params.values()))

                else:

                    def stratified_fn(params):
                        return func(tuple(params.values()))

        original_init["params"] = params_dict  # needs to be here, we need the params to be a dict for the serialization

        super().__init__(params=params_dict, func=stratified_fn, name=name, dtype=dtype, label=label)
        self.hs3.original_init.update(original_init)

    def __repr__(self):
        try:
            value = f"{self.numpy():.4g}" if tf.executing_eagerly() else "graph-node"
        except Exception:
            value = "ERROR OCCURRED"
        return f"<zfit.{self.__class__.__name__} '{self.name}' params={[(k, p.name) for k, p in self.params.items()]} value={value}>"


class ComposedParameterRepr(BaseRepr):
    _implementation = ComposedParameter
    _constructor = pydantic.PrivateAttr(make_param_constructor(ComposedParameter))
    hs3_type: Literal["ComposedParameter"] = pydantic.Field("ComposedParameter", alias="type")
    name: str
    func: str
    params: dict[str, Serializer.types.ParamTypeDiscriminated]
    unpack_params: Optional[bool]
    internal_params: Optional[
        Union[
            Serializer.types.ParamTypeDiscriminated,
            list[Serializer.types.ParamTypeDiscriminated],
            dict[str, Serializer.types.ParamTypeDiscriminated],
        ]
    ]

    @validator("func", pre=True)
    def _validate_value_pre(cls, value):
        if cls.orm_mode(value):
            value = dill.dumps(value).hex()
        return value

    @pydantic.root_validator(pre=True)
    def validate_all_functor(cls, values):
        if cls.orm_mode(values):
            values = values["hs3"].original_init
        return values

    def _to_orm(self, init):
        init = copy.copy(init)
        func = init.pop("func")
        params = init.pop("internal_params")

        init["params"] = params
        init["func"] = dill.loads(bytes.fromhex(func))
        return super()._to_orm(init)


class ComplexParameter(ComposedParameter):  # TODO: change to real, imag as input?
    @deprecated_args(None, "Use `func` instead.", "value_fn")
    def __init__(
        self,
        name: str,
        value_fn: Callable | None = None,
        *,
        func: Callable | None = None,
        params,
        dtype=ztypes.complex,
        label: str | None = None,
    ):
        """Create a complex parameter.

        .. note::
            Use the constructor class methods instead of the __init__() constructor:

            - :py:meth:`ComplexParameter.from_cartesian`
            - :py:meth:`ComplexParameter.from_polar`

        Args:
            name: Name of the parameter.
            func: Function that returns the value of the complex parameter and takes as arguments the real and
                imaginary part.
            params: List of the real and imaginary part of the complex parameter.
            dtype: Data type of the complex parameter.
            label: |@doc:param.init.label||@docend:param.init.label|
        """
        # legacy
        if value_fn is not None:
            if func is not None:
                msg = "Cannot specify both `value_fn` and `func`."
                raise ValueError(msg)
            func = value_fn
            del value_fn
        super().__init__(name, func=func, params=params, dtype=dtype, label=label)
        self._conj = None
        self._mod = None
        self._arg = None
        self._imag = None
        self._real = None

    @classmethod
    def from_cartesian(
        cls,
        name: str,
        real: ztyping.NumericalScalarType,
        imag: ztyping.NumericalScalarType,
        floating: bool = True,
        *,
        dtype=ztypes.complex,
        label: str | None = None,
    ) -> ComplexParameter:  # TODO: correct dtype handling, also below
        """Create a complex parameter from cartesian coordinates.

        Args:
            name: Name of the parameter.
            real: Real part of the complex number.
            imag: Imaginary part of the complex number.
            floating: If True, the parameter is floating. If False, the parameter is constant.
            dtype: Data type of the complex parameter.
            label: |@doc:param.init.label||@docend:param.init.label|
        """
        real = convert_to_parameter(real, name=name + "_real", prefer_constant=not floating)
        imag = convert_to_parameter(imag, name=name + "_imag", prefer_constant=not floating)
        param = cls(
            name=name,
            func=lambda _real, _imag: znp.asarray(tf.complex(_real, _imag), dtype=dtype),
            params=[real, imag],
            label=label,
        )
        param._real = real
        param._imag = imag
        return param

    @classmethod
    def from_polar(
        cls,
        name: str,
        mod: ztyping.NumericalScalarType,
        arg: ztyping.NumericalScalarType,
        floating=True,
        *,
        dtype=ztypes.complex,
        label: str | None = None,
        **__,
    ) -> ComplexParameter:
        """Create a complex parameter from polar coordinates.

        Args:
            name: Name of the parameter.
            mod: Modulus (r) the complex number.
            arg: Argument (phi) of the complex number.
            dtype: Data type of the complex parameter.
            floating: If True, the parameter is floating. If False, the parameter is constant.
            label: |@doc:param.init.label||@docend:param.init.label|
        """
        mod = convert_to_parameter(mod, name=name + "_mod", prefer_constant=not floating)
        arg = convert_to_parameter(arg, name=name + "_arg", prefer_constant=not floating)
        param = cls(
            name=name,
            func=lambda _mod, _arg: znp.asarray(tf.complex(_mod * znp.cos(_arg), _mod * znp.sin(_arg)), dtype=dtype),
            params=[mod, arg],
            label=label,
        )
        param._mod = mod
        param._arg = arg
        return param

    @property
    def conj(self):
        """Returns a complex conjugated copy of the complex parameter."""

        if self._conj is None:
            name = f"{self.name}_conj"
            if (label := self._label) is not None:
                label = f"Conjugate of {self.label}"
            self._conj = ComplexParameter(
                name=name,
                func=lambda: znp.conj(self),
                params=self.get_cache_deps(),
                dtype=self.dtype,
                label=label,
            )
        return self._conj

    @property
    def real(self) -> tf.Tensor:
        """Real part of the complex parameter."""
        return znp.real(self)

    @property
    def imag(self) -> tf.Tensor:
        """Imaginary part of the complex parameter."""
        return znp.imag(self)

    @property
    def mod(self) -> tf.Tensor:
        """Modulus (r) of the complex parameter."""
        return znp.abs(self)

    @property
    def arg(self) -> tf.Tensor:
        """Argument (phi) of the complex parameter."""
        return znp.angle(self)


# register_tensor_conversion(ConstantParameter, "ConstantParameter", True)
register_tensor_conversion(ComposedParameter, "ComposedParameter", True)

_auto_number = 0


def get_auto_number():
    global _auto_number
    auto_number = _auto_number
    _auto_number += 1
    return auto_number


def _reset_auto_number():
    global _auto_number
    _auto_number = 0


def convert_to_parameters(
    value,
    name: str | list[str] | None = None,
    prefer_constant: bool | None = None,
    lower=None,
    upper=None,
    step_size=None,
):
    if prefer_constant is None:
        prefer_constant = True
    if isinstance(value, collections.abc.Mapping):
        return convert_to_parameters(**value, prefer_constant=False)
    value = convert_to_container(value)
    is_param_already = [isinstance(val, ZfitIndependentParameter) for val in value]
    if all(is_param_already):
        return value
    elif any(is_param_already):
        msg = f"value has to be either ZfitParameters or values, not mixed (currently)." f" Is {value}."
        raise ValueError(msg)
    params_dict = {
        "value": value,
        "name": name,
        "lower": lower,
        "upper": upper,
        "step_size": step_size,
    }
    params_dict = {
        key: convert_to_container(val, ignore=np.ndarray) for key, val in params_dict.items() if val is not None
    }
    lengths = {len(v) for v in params_dict.values()}
    if len(lengths) != 1:
        msg = f"Inconsistent length in values when converting the parameters: {params_dict}"
        raise ValueError(msg)

    params = []
    for i in range(len(params_dict["value"])):
        pdict = {k: params_dict[k][i] for k in params_dict}
        params.append(convert_to_parameter(**pdict, prefer_constant=prefer_constant))
    return params


@deprecated_args(None, "Use `params` instead.", "dependents")
def convert_to_parameter(
    value: ztyping.NumericalScalarType | ZfitParameter | Callable,
    name: str | None = None,
    prefer_constant: bool = True,
    params: ZfitParameter | Iterable[ZfitParameter] | None = None,
    lower: ztyping.NumericalScalarType | None = None,
    upper: ztyping.NumericalScalarType | None = None,
    step_size: ztyping.NumericalScalarType | None = None,
    *,
    label: str | None = None,
    # legacy
    dependents=None,
) -> ZfitParameter:
    """Convert a *numerical* to a constant/floating parameter or return if already a parameter.

    Args:
        value: Value of the parameter. If a `ZfitParameter` is passed, it will be returned as is.
        name: Name of the parameter. If None, a unique name will be created.
        prefer_constant: If True, create a ConstantParameter instead of a Parameter, if possible.
        params: If the value is a callable, the parameters that are passed to the callable.
        lower: Lower limit of the parameter.
        upper: Upper limit of the parameter.
        step_size: Step size of the parameter.
        label: |@doc:param.init.label||@docend:param.init.label|
    """
    # legacy start
    if dependents is not None:
        params = dependents
    # legacy end
    if name is not None:
        name = str(name)

    if callable(value):
        if params is None:
            msg = "If the value is a callable, the params have to be specified as an empty list/tuple"
            raise ValueError(msg)
        return ComposedParameter(f"Composed_autoparam_{get_auto_number()}", func=value, params=params, label=label)

    if isinstance(value, ZfitParameter):  # TODO(Mayou36): autoconvert variable. TF 2.0?
        return value
    elif isinstance(value, tf.Variable):
        msg = f"Currently, cannot autoconvert tf.Variable ({value}) to zfit.Parameter."
        raise TypeError(msg)

    # convert to Tensor
    if not isinstance(value, tf.Tensor):
        value = z.to_complex(value) if isinstance(value, complex) else z.to_real(value)

    if not run._enable_parameter_autoconversion:
        return value

    if value.dtype.is_complex:
        if name is None:
            name = "FIXED_complex_autoparam_" + str(get_auto_number())
        if prefer_constant:
            complex_params = (
                ConstantParameter(name + "_REALPART", value=znp.real(value)),
                ConstantParameter(name + "_IMAGPART", value=znp.imag(value)),
            )
        else:
            complex_params = (
                Parameter(name + "_REALPART", value=znp.real(value)),
                Parameter(name + "_IMAGPART", value=znp.imag(value)),
            )
        value = ComplexParameter.from_cartesian(name, real=complex_params[0], imag=complex_params[1], label=label)

    elif prefer_constant:
        if name is None:
            name = "FIXED_autoparam_" + str(get_auto_number()) if name is None else name
        value = ConstantParameter(name, value=value, label=label)

    else:
        name = "autoparam_" + str(get_auto_number()) if name is None else name
        value = Parameter(name=name, value=value, lower=lower, upper=upper, step_size=step_size, label=label)

    return value


@z.function(wraps="params", keepalive=True)
def assign_values_jit(
    params: Parameter | Iterable[Parameter],
    values: ztyping.NumericalScalarType | Iterable[ztyping.NumericalScalarType],
    use_locking=False,
):
    """Assign values to parameters jitted.

    This method can be significantly faster than `set_values`, however it expects the correct data-type and
    cannot, for example, take a `FitResult` as input or function as a context manager. Only use when
    performance is critical (such as inside a minimizer).

    Args:
        params: The parameters to assign the values to.
        values: Values to assign to the parameters.
        use_locking:
    """
    for i, param in enumerate(params):
        value = values[i]
        if value.dtype != param.dtype:
            value = znp.cast(value, param.dtype)
        param.assign(value, read_value=False, use_locking=use_locking)


def assign_values(
    params: Parameter | Iterable[Parameter],
    values: ztyping.NumericalScalarType | Iterable[ztyping.NumericalScalarType],
    use_locking=False,
    allow_partial: bool | None = None,
):
    """Set the values of multiple parameters in a fast way.

    In general, :meth:`set_values` is to be preferred. `assign_values` will ignore out-of-bounds errors,
     does not offer a context-manager but is in general (an order of magnitude) faster.

    Args:
        params: Parameters to set the values.
        values: List-like object that supports indexing.
        use_locking: if true, lock the parameter to avoid race conditions.
        allow_partial: Allow to set only parts of the parameters in case values is a `ZfitResult`
            and not all are present in the
            *values*. If False, *params* not in *values* will raise an error.
            Note that setting this to true will also go with an empty values container.

    Raises:
        ValueError: If not all *params* are in *values* if *values* is a `FitResult` and `allow_partial` is `False`.
    """
    if allow_partial is None:
        allow_partial = False
    params, values, _ = check_convert_param_values_assign(params, values, allow_partial=allow_partial)
    assign_values_jit(params=params, values=values, use_locking=use_locking)


def set_values(
    params: Parameter | Iterable[Parameter],
    values: (
        ztyping.NumericalScalarType
        | Iterable[ztyping.NumericalScalarType]
        | ZfitResult
        | Mapping[str | ZfitParameter, ztyping.NumericalScalarType]
        | None,
    ) = None,
    allow_partial: bool | None = None,
):
    """Set the values (using a context manager or not) of multiple parameters.

    Args:
        params: Parameters to set the values.
        values: List-like object that supports indexing, a `FitResult` or a `Mapping` of parameters or parameter names to values.
        allow_partial: Allow to set only parts of the parameters in case values is a `ZfitResult`
            and not all are present in the
            *values*. If False, *params* not in *values* will raise an error.
            Note that setting this to true will also go with an empty values container.

    Returns:
        An object for a context manager (but can also be used without), can be ignored.
    Raises:
        ValueError: If the value is not between the limits of the parameter.
        ValueError: If not all *params* are in *values* if *values* is a `FitResult` and `allow_partial` is `False`.
    """
    if allow_partial is None:
        allow_partial = False
    params, values, _ = check_convert_param_values_assign(params, values, allow_partial)

    def setter(values):
        for i, param in enumerate(params):
            param.set_value(values[i])

    def getter():
        return [param.value() for param in params]

    return TemporarilySet(values, setter=setter, getter=getter)


def check_convert_param_values_assign(params, values, allow_partial=False):
    """Check if params are valid and convert them if necessary to be used with assign_values.

    Args:
        params: Parameters to set the values.
        values: List-like object that supports indexing or a `ZfitResult` or a `Mapping`.
        allow_partial: Allow to set only parts of the parameters in case values is a `ZfitResult` or a `Mapping`. Other parameters will not be updated.
            More values than parameters, if they are given in a `Mapping` or `ZfitResult`, are allowed in any case.
    Returns:
        A tuple of (params, values)
    """
    if isinstance(params, ZfitResult) and values is None:
        params, values = None, params
    elif isinstance(params, Mapping):
        if values is not None:
            msg = "Cannot set values to parameters if params are a Mapping."
            raise ValueError(msg)
        params, values = tuple(params.keys()), tuple(params.values())
    elif not isinstance(values, ZfitResult):
        params = convert_to_container(params)
        if params is None:
            msg = "No parameters given to set values to (values={values})."
            raise ValueError(msg)
        noparams = len(params) == 0
        if noparams:
            return params, values, True
    if isinstance(values, ZfitResult):
        result = values
        new_params = []
        values = []
        if params is None:
            params = result.params.keys()
        for param in params:
            if param in result.params:
                values.append(result.params[param]["value"])
                new_params.append(param)
            elif not allow_partial:
                msg = (
                    f"Cannot set {param} with {result} as it is not contained. To partially set"
                    f" the parameters (only those in the result), use allow_partial"
                )
                raise ValueError(msg)
        params = new_params
    elif isinstance(values, Mapping):
        new_params = []
        new_values = []
        for param in params:
            if param in values:
                new_values.append(values[param])
                new_params.append(param)
            elif param.name in values:
                new_values.append(values[param.name])
                new_params.append(param)
            elif not allow_partial:
                msg = (
                    f"Cannot set {param} with {values!r} as it is not contained. To partially set"
                    f" the parameters (only those in the result), use allow_partial"
                )
                raise ValueError(msg)
        values = new_values
        params = new_params

    else:
        if not (tf.is_tensor(values) or isinstance(values, np.ndarray)):
            values = convert_to_container(values)
            lenvalues = len(values)
        else:
            values = znp.asarray(values, dtype=znp.float64)
            values = znp.atleast_1d(values)
            shape = values.shape
            lenvalues = shape[0] if shape is not None else None
        if lenvalues is not None and lenvalues != len(params):
            msg = f"Incompatible length of parameters and values: {params}, {values}"
            raise ValueError(msg)

    not_param = [param for param in params if not isinstance(param, ZfitParameter)]
    if not_param:
        msg = f"The following are not parameters (but should be): {not_param}"
        raise TypeError(msg)
    non_independent_params = [param for param in params if not param.independent]
    if non_independent_params:
        msg = f"trying to set value of parameters that are not independent " f"{non_independent_params}"
        raise ParameterNotIndependentError(msg)
    values = znp.asarray(values, dtype=znp.float64)
    return params, values, False  # if it's empty



#  Copyright (c) 2024 zfit

from __future__ import annotations

import abc
import collections
from collections.abc import Callable
from typing import Iterable, Literal, Mapping, Optional

import numpy as np
import pydantic
import tensorflow as tf
import tensorflow_probability as tfp
from ordered_set import OrderedSet

import zfit.z.numpy as znp

from .. import z
from ..serialization.serializer import BaseRepr, Serializer
from ..settings import ztypes
from ..util import ztyping
from ..util.container import convert_to_container
from ..util.deprecation import deprecated_args
from ..util.exception import ShapeIncompatibleError
from .baseobject import BaseNumeric
from .dependents import _extract_dependencies
from .interfaces import ZfitConstraint, ZfitParameter
from .serialmixin import SerializableMixin

tfd = tfp.distributions


class BaseConstraintRepr(BaseRepr):
    _implementation = None
    _owndict = pydantic.PrivateAttr(default_factory=dict)
    hs3_type: Literal["BaseConstraint"] = pydantic.Field("BaseConstraint", alias="type")


class BaseConstraint(ZfitConstraint, BaseNumeric):
    def __init__(
        self,
        params: dict[str, ZfitParameter] | None = None,
        name: str = "BaseConstraint",
        dtype=ztypes.float,
        **kwargs,
    ):
        """Base class for constraints.

        Args:
            dtype: the dtype of the constraint
            name: the name of the constraint
            params: A dictionary with the internal name of the
                parameter and the parameters itself the constrains depends on
        """
        super().__init__(name=name, dtype=dtype, params=params, **kwargs)

    def value(self):
        return self._value()

    @abc.abstractmethod
    def _value(self):
        raise NotImplementedError

    def _get_dependencies(self) -> ztyping.DependentsType:
        return _extract_dependencies(self.get_params(floating=None))


# TODO: improve arbitrary constraints, should we allow only functions that have a `params` argument?
class SimpleConstraint(BaseConstraint):
    def __init__(
        self,
        func: Callable,
        params: (Mapping[str, ztyping.ParameterType] | Iterable[ztyping.ParameterType] | ztyping.ParameterType | None),
        *,
        name: str | None = None,
    ):
        """Constraint from a (function returning a) Tensor.

        Args:
            func: Callable that constructs the constraint and returns a tensor. For the expected signature,
                see below in ``params``.
            params: The parameters of the loss. If given as a list, the parameters are named "param_{i}"
                and the function does not take any arguments. If given as a dict, the function expects
                the parameter as the first argument (``params``).
        """
        if name is None:
            name = "SimpleConstraint"
        self._simple_func = func
        self._func_params = None
        if isinstance(params, collections.abc.Mapping):
            self._func_params = params
            params = list(params.values())
        self._simple_func_dependents = convert_to_container(params, container=OrderedSet)

        params = convert_to_container(params, container=list)
        params = {f"param_{i}": p for i, p in enumerate(params)} if self._func_params is None else self._func_params

        super().__init__(name=name, params=params)

    def _value(self):
        if self._func_params is None:
            return self._simple_func()
        else:
            return self._simple_func(self._func_params)


class ProbabilityConstraint(BaseConstraint):
    def __init__(
        self,
        observation: ztyping.NumericalScalarType | ZfitParameter,
        params: dict[str, ZfitParameter] | None = None,
        name: str = "ProbabilityConstraint",
        dtype=ztypes.float,
        **kwargs,
    ):
        """Base class for constraints using a probability density function.

        Args:
            dtype: the dtype of the constraint
            name: the name of the constraint
            params: The parameters to constraint
            observation: Observed values of the parameter
                to constraint obtained from auxiliary measurements.
        """
        # TODO: proper handling of input params, arrays. ArrayParam?
        if isinstance(params, collections.abc.Mapping):
            params_dict = params
            params = [p for name, p in params.items() if name.startswith("param_")]
        else:
            params = convert_to_container(params, ignore=np.ndarray, container=tuple)
            params_dict = {f"param_{i}": p for i, p in enumerate(params)}
        super().__init__(name=name, dtype=dtype, params=params_dict, **kwargs)

        observation = convert_to_container(observation, tuple, ignore=np.ndarray)
        if len(observation) != len(params):
            msg = (
                "observation and params have to be the same length. Currently"
                f"observation: {len(observation)}, params: {len(params)}"
            )
            raise ShapeIncompatibleError(msg)

        self._observation = observation  # TODO: needed below? Why?
        # for obs, p in zip(observation, params):
        #     obs = convert_to_parameter(obs, f"{p.name}_obs", prefer_constant=False)
        #     obs.floating = False
        #     self._observation.append(obs)

        self._ordered_params = params

    @property
    def observation(self):
        """Return the observed values of the parameters constrained."""
        return self._observation

    def value(self):
        return self._value()

    @abc.abstractmethod
    def _value(self):
        raise NotImplementedError

    def _get_dependencies(self) -> ztyping.DependentsType:
        return _extract_dependencies(self.get_params())

    def sample(self, n):
        """Sample ``n`` points from the probability density function for the observed value of the parameters.

        Args:
            n: The number of samples to be generated.
        Returns:
        """
        sample = self._sample(n=n)
        return {p: sample[:, i] for i, p in enumerate(self._ordered_params)}

    @abc.abstractmethod
    def _sample(self, n):
        raise NotImplementedError

    @property
    def _params_array(self):
        return znp.asarray(self._ordered_params)


class TFProbabilityConstraint(ProbabilityConstraint):
    def __init__(
        self,
        observation: ztyping.NumericalScalarType | ZfitParameter,
        params: dict[str, ZfitParameter],
        distribution: tfd.Distribution,
        dist_params,
        dist_kwargs=None,
        name: str = "DistributionConstraint",
        dtype=ztypes.float,
        **kwargs,
    ):
        """Base class for constraints using a probability density function from ``tensorflow_probability``.

        Args:
            distribution: The probability density function
                used to constraint the parameters
        """
        super().__init__(observation=observation, params=params, name=name, dtype=dtype, **kwargs)

        self._distribution = distribution
        self.dist_params = dist_params
        self.dist_kwargs = dist_kwargs if dist_kwargs is not None else {}

    @property
    def distribution(self):
        params = self.dist_params
        if callable(params):
            params = params(self.observation)
        kwargs = self.dist_kwargs
        if callable(kwargs):
            kwargs = kwargs()
        params = {k: znp.asarray(v, ztypes.float) for k, v in params.items()}
        return self._distribution(**params, **kwargs, name=f"{self.name}_tfp")

    def _value(self):
        array = znp.asarray(self._params_array, ztypes.float)
        value = -self.distribution.log_prob(array)
        return tf.reduce_sum(value)

    def _sample(self, n):
        return self.distribution.sample(n)


def _preprocess_gaussian_constr_sigma_var(cov, sigma, legacy_uncertainty):
    if sigma is not None:
        if legacy_uncertainty:
            msg = "Either `sigma` or `uncertainty` can be given, not both. Use `sigma`. `uncertainty` is deprecated."
            raise ValueError(msg)
        if cov is not None:
            msg = "Either `sigma` or `cov` can be given, not both."
            raise ValueError(msg)
        if any(isinstance(s, ZfitParameter) for s in convert_to_container(sigma)):
            msg = "sigma has to be a scalar or a 1D tensor, not a ZfitParameter (if this feature is needed, please open an issue on github with zfit."
            raise ValueError(msg)
        sigma = znp.asarray(sigma, ztypes.float)
        sigma = znp.atleast_1d(sigma)
        if (ndims := sigma.shape.ndims) == 2:
            msg = f"sigma has to be a scalar or a 1D tensor, not a {ndims}D tensor. Use `cov` instead."
            raise ValueError(msg)
        if ndims < 2:
            cov = znp.diag(znp.square(sigma))
        else:
            msg = f"sigma has to be a scalar, a 1D tensor or a 2D tensor, not {ndims}D."
            raise ValueError(msg)
    elif cov is not None:
        if any(isinstance(c, ZfitParameter) for c in convert_to_container(cov)):
            msg = "cov has to be a scalar, a 1D tensor or a 2D tensor, not a ZfitParameter (if this feature is needed, please open an issue on github with zfit."
            raise ValueError(msg)
        if legacy_uncertainty:
            msg = "Either `cov` or `uncertainty` can be given, not both. Use `cov`. `uncertainty` is deprecated."
            raise ValueError(msg)
        cov = znp.atleast_1d(znp.asarray(cov, ztypes.float))
        if cov.shape.ndims == 1:
            cov = znp.diag(cov)
        sigma = znp.sqrt(znp.diag(cov))
    else:  # legacy 3
        sigma = -999
        cov = -999
        # end legacy 3
    return sigma, cov


class GaussianConstraint(TFProbabilityConstraint, SerializableMixin):
    @deprecated_args(None, "Use `sigma` or `cov` instead.", "uncertainty")
    def __init__(
        self,
        params: ztyping.ParamTypeInput,
        observation: ztyping.NumericalScalarType,
        *,
        uncertainty: ztyping.NumericalScalarType = None,
        sigma: ztyping.NumericalScalarType = None,
        cov: ztyping.NumericalScalarType = None,
    ):
        r"""Gaussian constraints on a list of parameters to some observed values with uncertainties.

        A Gaussian constraint is defined as the likelihood of ``params`` given the ``observations`` and ``sigma`` or ``cov``
        from a different measurement.

        .. math::
            \text{constraint} = \text{Gauss}(\text{observation}; \text{params}, \text{uncertainty})


        Args:
            params: The parameters to constraint; corresponds to x in the Gaussian
                distribution.
            observation: observed values of the parameter; corresponds to mu
                in the Gaussian distribution.
            sigma: Typically the uncertainties of the observed values. Can either be a single value,
                a list of values, an array or a tensor. Must be broadcastable to the shape of the parameters.
                Either `sigma` or `cov` can be given, not both.
                ``sigma`` is the square root of the diagonal of the covariance matrix.
            cov: The covariance matrix of the observed values. Can either be a single value,
                a list of values, an array or a tensor that are either 1 or 2 dimensional. If 1D, it is interpreted
                as the diagonal of the covariance matrix.
                Either ``sigma`` or ``cov`` can be given, not both.
                ``cov`` is a 2D matrix with the shape `(n, n)` where `n` is the number of parameters and ``sigma``
                squared on the diagonal.
        Raises:
            ShapeIncompatibleError: If params, mu and sigma have incompatible shapes.
        """

        observation = convert_to_container(observation, tuple, ignore=np.ndarray)
        params = convert_to_container(params, tuple, ignore=np.ndarray)
        params_tuple_legacy = params

        # legacy start 1
        if legacy_uncertainty := uncertainty is not None:
            uncertainty = convert_to_container(uncertainty, tuple, ignore=np.ndarray)
            if isinstance(uncertainty[0], (np.ndarray, tf.Tensor)) and len(uncertainty) == 1:
                uncertainty = tuple(uncertainty[0])

            def create_covariance_legacy(mu, sigma):
                mu = z.convert_to_tensor(mu)
                sigma = znp.asarray(
                    sigma
                )  # otherwise TF complains that the shape got changed from [2] to [2, 2] (if we have a tuple of two arrays)
                sigma = z.convert_to_tensor(sigma)
                params_tensor = z.convert_to_tensor(params_tuple_legacy)

                if sigma.shape.ndims > 1:
                    covariance = sigma
                elif sigma.shape.ndims == 1:
                    covariance = tf.linalg.tensor_diag(z.pow(sigma, 2.0))
                else:
                    sigma = znp.reshape(sigma, [1])
                    covariance = tf.linalg.tensor_diag(z.pow(sigma, 2.0))

                if not params_tensor.shape[0] == mu.shape[0] == covariance.shape[0] == covariance.shape[1]:
                    msg = (
                        f"params_tensor, observation and uncertainty have to have the"
                        " same length. Currently"
                        f"param: {params_tensor.shape[0]}, mu: {mu.shape[0]}, "
                        f"covariance (from uncertainty): {covariance.shape[0:2]}"
                    )
                    raise ShapeIncompatibleError(msg)
                return covariance
        # legacy end 1

        original_init = {
            "observation": observation,
            "params": params,
            "uncertainty": uncertainty,
            "sigma": sigma,
            "cov": cov,
        }

        sigma, cov = _preprocess_gaussian_constr_sigma_var(cov, sigma, legacy_uncertainty)

        self.__cov = cov
        self.__sigma = sigma

        distribution = tfd.MultivariateNormalTriL

        def dist_params(observation, *, self=self):
            return {"loc": observation, "scale_tril": tf.linalg.cholesky(self.covariance)}

        dist_kwargs = {"validate_args": True}

        params = {f"param_{i}": p for i, p in enumerate(params)}

        super().__init__(
            name="GaussianConstraint",
            observation=observation,
            params=params,
            distribution=distribution,
            dist_params=dist_params,
            dist_kwargs=dist_kwargs,
        )
        self.hs3.original_init.update(original_init)
        if legacy_uncertainty:
            self._covariance = lambda: create_covariance_legacy(self.observation, uncertainty)
        else:
            self._covariance = lambda cov: znp.asarray(cov, ztypes.float)
        self._legacy_uncertainty = legacy_uncertainty

    @property
    def covariance(self):
        """Return the covariance matrix of the observed values of the parameters constrained."""
        # legacy start 2
        if self._legacy_uncertainty:
            return self._covariance()
        # legacy end 2
        return self._covariance(cov=self.__cov)


class GaussianConstraintRepr(BaseConstraintRepr):
    _implementation = GaussianConstraint
    hs3_type: Literal["GaussianConstraint"] = pydantic.Field("GaussianConstraint", alias="type")

    params: list[Serializer.types.ParamInputTypeDiscriminated]
    observation: list[Serializer.types.ParamInputTypeDiscriminated]
    uncertainty: Optional[list[Serializer.types.ParamInputTypeDiscriminated]]
    sigma: Optional[list[Serializer.types.ParamInputTypeDiscriminated]]
    cov: Optional[list[Serializer.types.ParamInputTypeDiscriminated]]

    @pydantic.root_validator(pre=True)
    def get_init_args(cls, values):
        if cls.orm_mode(values):
            values = values["hs3"].original_init
        return values

    @pydantic.validator("params", "observation", "uncertainty", "sigma", "cov")
    def validate_params(cls, v):
        return v.tolist() if isinstance(v, np.ndarray) else convert_to_container(v, list)


class PoissonConstraint(TFProbabilityConstraint, SerializableMixin):
    def __init__(self, params: ztyping.ParamTypeInput, observation: ztyping.NumericalScalarType):
        r"""Poisson constraints on a list of parameters to some observed values.

        Constraints parameters that can be counts (i.e. from a histogram) or, more generally, are
        Poisson distributed. This is often used in the case of histogram templates which are obtained
        from simulation and have a poisson uncertainty due to limited statistics.

        .. math::
            \text{constraint} = \text{Poisson}(\text{observation}; \text{params})


        Args:
            params: The parameters to constraint; corresponds to the mu in the Poisson
                distribution.
            observation: observed values of the parameter; corresponds to lambda
                in the Poisson distribution.
        Raises:
            ShapeIncompatibleError: If params and observation have incompatible shapes.
        """

        observation = convert_to_container(observation, tuple)
        params = convert_to_container(params, tuple)
        original_init = {"observation": observation, "params": params}

        distribution = tfd.Poisson
        dist_params = {"rate": observation}
        dist_kwargs = {"validate_args": False}

        super().__init__(
            name="PoissonConstraint",
            observation=observation,
            params=params,
            distribution=distribution,
            dist_params=dist_params,
            dist_kwargs=dist_kwargs,
        )
        self.hs3.original_init.update(original_init)


class PoissonConstraintRepr(BaseConstraintRepr):
    _implementation = PoissonConstraint
    hs3_type: Literal["PoissonConstraint"] = pydantic.Field("PoissonConstraint", alias="type")

    params: list[Serializer.types.ParamInputTypeDiscriminated]
    observation: list[Serializer.types.ParamInputTypeDiscriminated]

    @pydantic.root_validator(pre=True)
    def get_init_args(cls, values):
        if cls.orm_mode(values):
            values = values["hs3"].original_init
        return values

    @pydantic.validator("params", "observation")
    def validate_params(cls, v):
        return v.tolist() if isinstance(v, np.ndarray) else convert_to_container(v, list)


class LogNormalConstraint(TFProbabilityConstraint, SerializableMixin):
    def __init__(
        self,
        params: ztyping.ParamTypeInput,
        observation: ztyping.NumericalScalarType,
        uncertainty: ztyping.NumericalScalarType,
    ):
        r"""Log-normal constraints on a list of parameters to some observed values.

        Constraints parameters that can be counts (i.e. from a histogram) or, more generally, are
        LogNormal distributed. This is often used in the case of histogram templates which are obtained
        from simulation and have a log-normal uncertainty due to a multiplicative uncertainty.

        .. math::
            \text{constraint} = \text{LogNormal}(\text{observation}; \text{params})


        Args:
            params: The parameters to constraint; corresponds to the mu in the Poisson
                distribution.
            observation: observed values of the parameter; corresponds to lambda
                in the Poisson distribution.
            uncertainty: uncertainty of the observed values of the parameter; corresponds to sigma
                in the Poisson distribution.
        Raises:
            ShapeIncompatibleError: If params, mu and sigma have incompatible shapes.
        """

        observation = convert_to_container(observation, tuple)
        params = convert_to_container(params, tuple)
        uncertainty = convert_to_container(uncertainty, tuple)
        original_init = {
            "observation": observation,
            "params": params,
            "uncertainty": uncertainty,
        }

        distribution = tfd.LogNormal

        def dist_params(observation):
            return {"loc": observation, "scale": uncertainty}

        dist_kwargs = {"validate_args": False}

        super().__init__(
            name="LogNormalConstraint",
            observation=observation,
            params=params,
            distribution=distribution,
            dist_params=dist_params,
            dist_kwargs=dist_kwargs,
        )
        self.hs3.original_init.update(original_init)


class LogNormalConstraintRepr(BaseConstraintRepr):
    _implementation = LogNormalConstraint
    hs3_type: Literal["LogNormalConstraint"] = pydantic.Field("LogNormalConstraint", alias="type")

    params: list[Serializer.types.ParamInputTypeDiscriminated]
    observation: list[Serializer.types.ParamInputTypeDiscriminated]
    uncertainty: list[Serializer.types.ParamInputTypeDiscriminated]

    @pydantic.root_validator(pre=True)
    def get_init_args(cls, values):
        if cls.orm_mode(values):
            values = values["hs3"].original_init
        return values

    @pydantic.validator("params", "observation", "uncertainty")
    def validate_params(cls, v):
        return v.tolist() if isinstance(v, np.ndarray) else convert_to_container(v, list)



#  Copyright (c) 2024 zfit

from __future__ import annotations


def make_param_constructor(constructor):
    """Create a constructor for a parameter class avoiding the `NameAlreadyTakenError`.

    If the parameter already exists, it is returned instead of creating a new one.

    Args:
        constructor: Callable that creates the parameter.

    Returns:
        Callable that creates the parameter.
    """

    def param_constructor(name, **kwargs):
        from zfit.serialization import Serializer

        previous_existing = Serializer._existing_params

        if (param := previous_existing.get(name)) is None:
            Serializer._existing_params[name] = (param := constructor(name=name, **kwargs))
        return param

    return param_constructor



#  Copyright (c) 2024 zfit
from __future__ import annotations

import numpy as np

import zfit

# create space
obs = zfit.Space("x", -2, 3)

# parameters
mu = zfit.Parameter("mu", 1.2, -4, 6)
sigma = zfit.Parameter("sigma", 1.3, 0.5, 10)

# model building, pdf creation
gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs)

# data
data = np.random.normal(loc=2.0, scale=3.0, size=10000)
# we can convert it to a zfit Data object to make sure the data is within the limits or use the space to filter manually
data = obs.filter(data)  # works also for pandas DataFrame
# data = zfit.Data.from_numpy(obs=obs, array=data)

# create NLL
nll = zfit.loss.UnbinnedNLL(model=gauss, data=data)

# create a minimizer
minimizer = zfit.minimize.Minuit(gradient=False)
result = minimizer.minimize(nll).update_params()

# do the error calculations with a hessian approximation
param_errors = result.hesse()

# or here with minos
param_errors_asymmetric, new_result = result.errors()



#  Copyright (c) 2024 zfit

import zfit.z.numpy as znp

def test_type_casting():

    one = 1.0
    oneplus = one + 1e-15

    assert one != oneplus, "Test is flawed"
    assert znp.asarray(one) != znp.asarray(oneplus)
    assert znp.asarray(one, znp.float32) == znp.asarray(oneplus, znp.float32)



#  Copyright (c) 2024 zfit
import numpy as np
import pytest
import zfit
from zfit import z
import zfit.z.numpy as znp
from numba_stats import qgaussian as qgaussian_numba
import tensorflow as tf

mu_true = 0.5
sigma_true = 1.1
# obtained from the numba_stats package by numerically integrating the qgaussian PDF
true_integral_dict = {1.00001: 0.9999782, 1.5: 0.9695, 2.0: 0.8063152, 2.5: 0.4805167, 2.9: 0.1078687, 2.99: 0.011044}


def create_qgauss(q, mu, sigma, limits):
    obs = zfit.Space("obs1", limits)
    qgauss = zfit.pdf.QGauss(q=q, mu=mu, sigma=sigma, obs=obs)
    return qgauss, obs


@pytest.mark.parametrize("q", [1.00001, 1.5, 2.0, 2.5, 2.9, 2.99])
def test_qgauss_pdf(q):
    qgauss, _ = create_qgauss(q=q, mu=mu_true, sigma=sigma_true, limits=(-5, 5))
    assert pytest.approx(
        np.atleast_1d(qgaussian_numba.pdf(0.5, q=q, mu=mu_true, sigma=sigma_true)), rel=1e-5
    ) == qgauss.pdf(0.5, norm=False)
    test_values = znp.linspace(-5, 5, 10_000)
    np.testing.assert_allclose(
        qgauss.pdf(test_values, norm=False),
        qgaussian_numba.pdf(test_values, q=q, mu=mu_true, sigma=sigma_true),
        rtol=1e-5,
    )
    assert np.all(qgauss.pdf(test_values, norm=False) <= qgauss.pdf(0.5, norm=False))

    sample = qgauss.sample(1000)
    assert np.all(np.isfinite(sample.value())), "Some samples from the qgauss PDF are NaN or infinite"
    assert sample.n_events == 1000
    assert np.all(tf.logical_and(-5 <= sample.value(), sample.value() <= 5))


@pytest.mark.parametrize("q", [1.00001, 1.5, 2.0, 2.5, 2.9, 2.99])
def test_qgauss_integral(q):
    qgauss, obs = create_qgauss(q=q, mu=mu_true, sigma=sigma_true, limits=(-5, 5))
    full_interval_analytic = qgauss.analytic_integrate(obs, norm=False)
    full_interval_numeric = qgauss.numeric_integrate(obs, norm=False)
    true_integral = true_integral_dict[q]
    numba_stats_full_integral = qgaussian_numba.cdf(5, q=q, mu=mu_true, sigma=sigma_true) - qgaussian_numba.cdf(
        -5, q=q, mu=mu_true, sigma=sigma_true
    )
    assert pytest.approx(true_integral, 1e-4) == full_interval_analytic
    assert pytest.approx(true_integral, 1e-4) == full_interval_numeric
    assert pytest.approx(numba_stats_full_integral, 1e-6) == full_interval_analytic
    assert pytest.approx(numba_stats_full_integral, 1e-6) == full_interval_numeric

    analytic_integral = qgauss.analytic_integrate(limits=(-1, 1), norm=False)
    numeric_integral = qgauss.numeric_integrate(limits=(-1, 1), norm=False)
    numba_stats_integral = qgaussian_numba.cdf(1, q=q, mu=mu_true, sigma=sigma_true) - qgaussian_numba.cdf(
        -1, q=q, mu=mu_true, sigma=sigma_true
    )
    assert pytest.approx(numeric_integral, 1e-5) == analytic_integral
    assert pytest.approx(numba_stats_integral, 1e-5) == analytic_integral



#  Copyright (c) 2024 zfit
from __future__ import annotations

import tensorflow as tf

import zfit
from zfit import z


class CustomPDF2D(zfit.pdf.BasePDF):
    """My custom, 2 dimensional pdf where the axes are: Energy, Momentum."""

    def __init__(
        self,
        param1,
        param2,
        param3,
        obs,
        name="CustomPDF",
    ):
        # we can now do complicated stuff here if needed
        # only thing: we have to specify explicitly here what is which parameter
        params = {
            "super_param": param1,  # we can change/compose etc parameters
            "param2": param2,
            "param3": param3,
        }
        super().__init__(obs, params, name=name)

    @zfit.supports()
    def _unnormalized_pdf(self, x, params):
        energy = x[0]
        momentum = x[1]
        param1 = params["super_param"]
        param2 = params["param2"]
        param3 = params["param3"]

        # just a fantasy function
        return param1 * tf.cos(energy**2) + tf.math.log(param2 * momentum**2) + param3


# add an analytic integral


# define the integral function
def integral_full(limits, norm_range, params, model):
    del norm_range, model  # not used here
    lower, upper = limits.v1.limits
    param1 = params["super_param"]
    param2 = params["param2"]
    param3 = params["param3"]

    # calculate the integral here, dummy integral, wrong!
    return param1 * param2 * param3 + z.reduce_sum([lower, upper])


# define the space over which it is defined. Here, we use the axes
integlimits_axis0 = zfit.Space(axes=0, lower=-10, upper=10)
integlimits_axis1 = zfit.Space(axes=1, lower=zfit.Space.ANY_LOWER, upper=zfit.Space.ANY_UPPER)

integral_full_limits = integlimits_axis0 * integlimits_axis1  # creates 2D space
CustomPDF2D.register_analytic_integral(func=integral_full, limits=integral_full_limits)


# define the partial integral function
def integral_axis1(x, limits, norm_range, params, model):
    del norm_range, model  # not used here

    data_0 = x[0]  # data from axis 0

    param1 = params["super_param"]
    param2 = params["param2"]
    param3 = params["param3"]

    lower, upper = limits.limit1d  # for a more detailed guide, see the space.py example
    lower = z.convert_to_tensor(lower)  # the limits are now 1-D, for axis 1
    upper = z.convert_to_tensor(upper)

    # calculate the integral here, dummy integral
    return data_0**2 * param1 * param2 * param3 + z.reduce_sum([lower, upper])
    # notice that the returned shape will be in the same as data_0, e.g. the number of events given in x


# define the space over which it is defined. Here, we use the axes
integral_axis1_limits = zfit.Space(
    axes=(1,),  # axes one corresponds to the second obs, here obs2
    lower=zfit.Space.ANY_LOWER,
    upper=zfit.Space.ANY_UPPER,
)

CustomPDF2D.register_analytic_integral(func=integral_axis1, limits=integral_axis1_limits)

if __name__ == "__main__":
    import numpy as np

    obs = zfit.Space("obs1", -10, 10) * zfit.Space("obs2", -3, 5)
    pdf = CustomPDF2D(
        param1=1, param2=2, param3=3, obs=obs
    )  # if a Python number is passed, it will be regarded as a constant
    sample = pdf.sample(n=1000)
    pdf.pdf([[2.0, 2.5], [5.4, 3.2]])
    # x_part = zfit.Data(np.array([2.1, 2.2, 3.2]), obs="obs1")
    x_part = np.array([2.1, 2.2, 3.2])
    # integrate over obs2 with limits 1, 2 for the `x_part`. This will use the analytic integral above
    pdf.partial_integrate(x=x_part, limits=zfit.Space("obs2", 1, 2))
    # we can explicitly call the analytic integral. Without registering it (e.g. comment the line with the `register`
    # and run again), it will raise an error
    pdf.partial_analytic_integrate(x=x_part, limits=zfit.Space("obs2", 1, 2))



#  Copyright (c) 2024 zfit
from __future__ import annotations

from .minimizers.baseminimizer import (
    BaseMinimizer,
    BaseMinimizerV1,
    DefaultStrategy,
    minimize_supports,
)
from .minimizers.ipopt import IpyoptV1
from .minimizers.minimizer_minuit import Minuit
from .minimizers.minimizer_nlopt import (
    NLoptBaseMinimizerV1,
    NLoptBOBYQAV1,
    NLoptCCSAQV1,
    NLoptCOBYLAV1,
    NLoptESCHV1,
    NLoptISRESV1,
    NLoptLBFGSV1,
    NLoptMLSLV1,
    NLoptMMAV1,
    NLoptShiftVarV1,
    NLoptSLSQPV1,
    NLoptStoGOV1,
    NLoptSubplexV1,
    NLoptTruncNewtonV1,
)
from .minimizers.minimizers_scipy import (
    ScipyBaseMinimizerV1,
    ScipyLBFGSBV1,
    ScipyNelderMeadV1,
    ScipyNewtonCGV1,
    ScipyPowellV1,
    ScipySLSQPV1,
    ScipyTruncNCV1,
    ScipyTrustConstrV1,
)
from .minimizers.optimizers_tf import Adam, WrapOptimizer
from .minimizers.strategy import DefaultToyStrategy, PushbackStrategy
from .minimizers.termination import EDM

__all__ = [
    "WrapOptimizer",
    "Adam",
    "Minuit",
    "ScipyBaseMinimizerV1",
    "ScipyLBFGSBV1",
    "ScipyTrustConstrV1",
    "ScipyPowellV1",
    "ScipySLSQPV1",
    "ScipyNewtonCGV1",
    "ScipyTruncNCV1",
    "ScipyNelderMeadV1",
    "NLoptBaseMinimizerV1",
    "NLoptLBFGSV1",
    "NLoptTruncNewtonV1",
    "NLoptSLSQPV1",
    "NLoptMMAV1",
    "NLoptCCSAQV1",
    "NLoptShiftVarV1",
    "NLoptMLSLV1",
    "NLoptStoGOV1",
    "NLoptESCHV1",
    "NLoptISRESV1",
    "NLoptSubplexV1",
    "NLoptBOBYQAV1",
    "NLoptCOBYLAV1",
    "IpyoptV1",
    "BaseMinimizer",
    "BaseMinimizerV1",
    "minimize_supports",
    "DefaultStrategy",
    "DefaultToyStrategy",
    "PushbackStrategy",
    "EDM",
]



#  Copyright (c) 2024 zfit
import copy

import numpy as np
import pytest
import tensorflow as tf
from numba_stats import bernstein as bernstein_numba

import zfit
from zfit import z
import zfit.z.numpy as znp

obs1_random = zfit.Space(obs="obs1", limits=(-1.5, 1.2))
obs1 = zfit.Space(obs="obs1", limits=(-1, 1))
obs2_random = zfit.Space(obs="obs2", limits=(0.5, 1.8))
obs2 = zfit.Space(obs="obs2", limits=(0, 1))

coeffs_parametrization = [
    1.4,
    [0.6],
    [1.42, 1.2],
    [0.2, 0.8, 0.5],
    [8.1, 1.4, 3.6, 4.1],
    [1.1, 1.42, 1.2, 0.4, 0.7],
    [11.1, 1.4, 5.6, 3.1, 18.1, 3.1],
]

rel_integral = 7e-2
default_sampling = 100000

poly_pdfs = [
    (zfit.pdf.Legendre, default_sampling),
    (zfit.pdf.Chebyshev, default_sampling),
    (zfit.pdf.Chebyshev2, default_sampling),
    (zfit.pdf.Hermite, default_sampling * 20),
    (zfit.pdf.Laguerre, default_sampling * 20),
    (zfit.pdf.Bernstein, default_sampling),
]


@pytest.mark.parametrize("poly_cfg", poly_pdfs)
@pytest.mark.parametrize("coeffs", coeffs_parametrization)
@pytest.mark.flaky(3)
def test_polynomials(poly_cfg, coeffs):
    coeffs = copy.copy(coeffs)
    poly_pdf, n_sampling = poly_cfg
    polynomial = poly_pdf(obs=obs1, coeffs=coeffs)
    polynomial2 = poly_pdf(obs=obs1, coeffs=coeffs)

    if poly_pdf == zfit.pdf.Bernstein:
        polynomial_coeff0 = poly_pdf(obs=obs1, coeffs=coeffs)
    else:
        polynomial_coeff0 = poly_pdf(obs=obs1, coeffs=coeffs, coeff0=1.0)
    lower, upper = obs1.v1.limits
    x = np.random.uniform(size=(1000,), low=lower, high=upper)
    y_poly = polynomial.pdf(x)
    y_poly_u = polynomial.pdf(x, norm=False)
    y_poly2 = polynomial2.pdf(x)
    y_poly2_u = polynomial2.pdf(x, norm=False)
    y_poly_coeff0 = polynomial_coeff0.pdf(x)
    y_poly_coeff0_u = polynomial_coeff0.pdf(x, norm=False)
    y_poly_np, y_poly2_np, y_poly_coeff0_np = [
        y_poly,
        y_poly2,
        y_poly_coeff0,
    ]
    y_polyu_np, y_poly2u_np, y_polyu_coeff0_np = [
        y_poly_u,
        y_poly2_u,
        y_poly_coeff0_u,
    ]
    np.testing.assert_allclose(y_polyu_np, y_poly2u_np)
    np.testing.assert_allclose(y_polyu_np, y_polyu_coeff0_np)
    np.testing.assert_allclose(y_poly_np, y_poly2_np)
    np.testing.assert_allclose(y_poly_np, y_poly_coeff0_np)

    # test 1 to 1 range
    polynomial = poly_pdf(obs=obs1, coeffs=coeffs)
    integral = polynomial.analytic_integrate(limits=obs1, norm=False)
    numerical_integral = polynomial.numeric_integrate(limits=obs1, norm=False)
    analytic_integral = integral
    assert (
        pytest.approx(analytic_integral, rel=rel_integral) == numerical_integral
    )

    # test 0 to 1 range
    polynomial = poly_pdf(obs=obs2, coeffs=coeffs)
    integral = polynomial.analytic_integrate(limits=obs2, norm=False)
    numerical_integral = polynomial.numeric_integrate(limits=obs2, norm=False)
    analytic_integral = integral
    assert (
        pytest.approx(analytic_integral, rel=rel_integral) == numerical_integral
    )

    # test with different range scaling, 1 to 1
    polynomial = poly_pdf(obs=obs1_random, coeffs=coeffs)

    # test with limits != space
    integral = polynomial.analytic_integrate(limits=obs1, norm=False)
    numerical_integral = polynomial.numeric_integrate(limits=obs1, norm=False)
    analytic_integral = integral
    assert (
        pytest.approx(analytic_integral, rel=rel_integral) == numerical_integral
    )

    # test with limits == space
    integral = polynomial.analytic_integrate(limits=obs1_random, norm=False)
    numerical_integral = polynomial.numeric_integrate(limits=obs1_random, norm=False)
    analytic_integral = integral
    assert (
        pytest.approx(analytic_integral, rel=rel_integral) == numerical_integral
    )

    lower, upper = obs1_random.limit1d
    sample = z.random.uniform((n_sampling, 1), lower, upper, dtype=tf.float64)
    test_integral = (
        np.average(polynomial.pdf(sample, norm=False)) * obs1_random.volume
    )
    assert pytest.approx(analytic_integral, rel=rel_integral * 3) == test_integral

    # test with different range scaling, 0 to 1
    polynomial = poly_pdf(obs=obs2_random, coeffs=coeffs)

    # test with limits != space
    integral = polynomial.analytic_integrate(limits=obs2, norm=False)
    numerical_integral = polynomial.numeric_integrate(limits=obs2, norm=False)
    analytic_integral = integral
    assert (
        pytest.approx(analytic_integral, rel=rel_integral) == numerical_integral
    )

    # test with limits == space
    integral = polynomial.analytic_integrate(limits=obs2_random, norm=False)
    numerical_integral = polynomial.numeric_integrate(limits=obs2_random, norm=False)
    analytic_integral = integral
    assert (
        pytest.approx(analytic_integral, rel=rel_integral) == numerical_integral
    )

    lower, upper = obs2_random.limit1d
    sample = z.random.uniform((n_sampling, 1), lower, upper, dtype=tf.float64)
    test_integral = (
        np.average(polynomial.pdf(sample, norm=False)) * obs2_random.volume
     )
    assert pytest.approx(analytic_integral, rel=rel_integral * 3) == test_integral




@pytest.mark.parametrize("coeffs", coeffs_parametrization)
@pytest.mark.parametrize("obs", [obs1, obs2, obs1_random, obs2_random])
def test_bernstein(coeffs, obs):
    zfit.run.set_graph_mode(False)
    bernstein = zfit.pdf.Bernstein(obs=obs, coeffs=coeffs)
    lower, upper = obs.limit1d

    assert pytest.approx(
        np.atleast_1d(bernstein_numba.density(0.8, beta=coeffs, xmin=lower, xmax=upper)), rel=1e-5
    ) == bernstein.pdf(0.8, norm=False)
    test_values = znp.linspace(lower, upper, 10_000)
    np.testing.assert_allclose(
        bernstein.pdf(test_values, norm=False),
        bernstein_numba.density(test_values, beta=coeffs, xmin=lower, xmax=upper),
        rtol=1e-5,
    )

    sample = bernstein.sample(1000)
    assert all(np.isfinite(sample.value())), "Some samples from the Bernstein PDF are NaN or infinite"
    assert sample.n_events == 1000
    assert all(tf.logical_and(lower <= sample.value(), sample.value() <= upper))


    full_interval_analytic = bernstein.analytic_integrate(obs, norm=False)
    full_interval_numeric = bernstein.numeric_integrate(obs, norm=False)
    numba_stats_full_integral = bernstein_numba.integral(x=upper, beta=coeffs, xmin=lower, xmax=upper) - bernstein_numba.integral(
        x=lower, beta=coeffs, xmin=lower, xmax=upper
    )
    assert pytest.approx(full_interval_numeric, 1e-4) == full_interval_analytic
    assert pytest.approx(numba_stats_full_integral, 1e-6) == full_interval_analytic
    assert pytest.approx(numba_stats_full_integral, 1e-6) == full_interval_numeric

    analytic_integral = bernstein.analytic_integrate(limits=(0.6, 0.9), norm=False)
    numeric_integral = bernstein.numeric_integrate(limits=(0.6, 0.9), norm=False)
    numba_stats_integral = bernstein_numba.integral(x=0.9, beta=coeffs, xmin=lower, xmax=upper) - bernstein_numba.integral(
        x=0.6, beta=coeffs, xmin=lower, xmax=upper
    )
    assert pytest.approx(numeric_integral, 1e-5) == analytic_integral
    assert pytest.approx(numba_stats_integral, 1e-5) == analytic_integral
    assert pytest.approx(numba_stats_integral, 1e-5) == numeric_integral



#  Copyright (c) 2024 zfit
from __future__ import annotations

from .models.functions import BaseFuncV1, ProdFunc, SimpleFuncV1, SumFunc, ZFuncV1

__all__ = ["BaseFuncV1", "ProdFunc", "SumFunc", "SimpleFuncV1", "ZFuncV1"]



"""The setup script."""

#  Copyright (c) 2024 zfit
from __future__ import annotations

import functools
import operator
import platform
import warnings
from pathlib import Path

from setuptools import setup

here = Path(__file__).resolve().parent

with Path(here / "requirements.txt", encoding="utf-8").open() as requirements_file:
    requirements = requirements_file.read().splitlines()

with Path(here / "requirements_dev.txt", encoding="utf-8").open() as requirements_dev_file:
    requirements_dev = requirements_dev_file.read().splitlines()
requirements_dev = [req.split("#")[0].strip() for req in requirements_dev]

nlopt_req = "nlopt>=2.7.1"
ipyopt_req = "ipyopt>=0.12"
extras_require = {
    "ipyopt": [ipyopt_req],
    "nlopt": [nlopt_req],
    "hs3": ["asdf"],
    "uproot": ["awkward-pandas"],
}
allreq = functools.reduce(operator.iadd, extras_require.values(), [])

tests_require = [
    "pytest>=3.4.2",  # breaks unittests
    "pytest-runner>=2.11.1",
    "pytest-rerunfailures>=6",
    "pytest-xdist",
    "pytest-ordering",
    "pytest-randomly",
    "scikit-hep-testdata",
    "pytest-timeout>=1",
    "matplotlib",  # for plots in examples
]
allreq_nonloptipyopt = allreq.copy()
allreq_nonloptipyopt.pop(allreq.index(nlopt_req))
allreq_nonloptipyopt.pop(allreq.index(ipyopt_req))

extras_require["all-linux"] = allreq
extras_require["all-darwin"] = allreq_nonloptipyopt + extras_require["nlopt"]
extras_require["all-windows"] = allreq_nonloptipyopt + extras_require["nlopt"]
extras_require["all-silicon"] = allreq_nonloptipyopt

extras_require["tests-linux"] = [*tests_require, nlopt_req, ipyopt_req]
extras_require["tests-darwin"] = [*tests_require, nlopt_req]
extras_require["tests-windows"] = [*tests_require, nlopt_req]
extras_require["tests-silicon"] = tests_require

extras_require["dev-linux"] = requirements_dev + extras_require["tests-linux"]
extras_require["dev-darwin"] = requirements_dev + extras_require["tests-darwin"]
extras_require["dev-windows"] = requirements_dev + extras_require["tests-windows"]
extras_require["dev-silicon"] = requirements_dev + extras_require["tests-silicon"]

extras_require["alldev-linux"] = extras_require["all-linux"] + extras_require["dev-linux"]
extras_require["alldev-darwin"] = extras_require["all-darwin"] + extras_require["dev-darwin"]
extras_require["alldev-silicon"] = extras_require["all-silicon"] + extras_require["dev-silicon"]

alldev_windows = extras_require["all-windows"] + extras_require["dev-windows"]
# not available on Windows: https://github.com/google/jax/issues/438#issuecomment-939866186
alldev_windows.pop(alldev_windows.index("jaxlib"))
extras_require["alldev-windows"] = alldev_windows

# fill defaults depending on the system
if (platf := platform.system().lower()) == "darwin" and platform.processor() == "arm":
    platf = "silicon"
if platf not in ["linux", "windows", "silicon", "darwin"]:
    warnings.warn(
        f"Platform {platf} not recognized, `dev`, `tests` and `alldev` extras contain all requirements. ",
        UserWarning,
        stacklevel=1,
    )
    platf = "linux"
extras_require["tests"] = extras_require[f"tests-{platf}"]
extras_require["dev"] = extras_require[f"dev-{platf}"]
extras_require["alldev"] = extras_require[f"alldev-{platf}"]
extras_require["all"] = extras_require[f"all-{platf}"]

cleaned_req = {}

for req_name, req in extras_require.items():
    req = list(set(req))
    req = [r.split("#")[0].strip() for r in req]
    req = [r for r in req if r]  # remove empty string
    cleaned_req[req_name] = req

setup(
    install_requires=requirements,
    tests_require=tests_require,
    extras_require=cleaned_req,
    use_scm_version=True,
)



#  Copyright (c) 2024 zfit
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import zfit

import tensorflow as tf

import zfit
import zfit.z.numpy as znp
from zfit import z

from ..core.interfaces import ZfitPDF, ZfitSpace
from ..util import ztyping
from ..util.warnings import warn_advanced_feature
from .binned_functor import BaseBinnedFunctorPDF


class MapNotVectorized(Exception):
    pass


class BinnedFromUnbinnedPDF(BaseBinnedFunctorPDF):
    def __init__(
        self,
        pdf: ZfitPDF,
        space: ZfitSpace,
        *,
        extended: ztyping.ExtendedInputType = None,
        norm: ztyping.NormInputType = None,
        name: str | None = None,
        label: str | None = None,
    ) -> None:
        """Create a binned pdf from an unbinned pdf binning in *space*.

        Args:
            pdf: The unbinned pdf to be binned.
            space: |@doc:pdf.init.obs| Observables of the
               model. This will be used as the default space of the PDF and,
               if not given explicitly, as the normalization range.

               The default space is used for example in the sample method: if no
               sampling limits are given, the default space is used.

               If the observables are binned and the model is unbinned, the
               model will be a binned model, by wrapping the model in a
               :py:class:`~zfit.pdf.BinnedFromUnbinnedPDF`, equivalent to
               calling :py:meth:`~zfit.pdf.BasePDF.to_binned`.

               The observables are not equal to the domain as it does not restrict or
               truncate the model outside this range. |@docend:pdf.init.obs|
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.name| Name of the PDF.
               Maybe has implications on the serialization and deserialization of the PDF.
               For a human-readable name, use the label. |@docend:pdf.init.name|
            label: |@doc:pdf.init.label| Human-readable name
               or label of
               the PDF for a better description, to be used with plots etc.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.label|
        """
        if pdf.is_extended:
            if extended is not None:
                warn_advanced_feature(
                    f"PDF {pdf} is already extended, but extended also given {extended}. Will" f" use the given yield.",
                    identifier="extend_wrapped_extended",
                )
            else:
                extended = pdf.get_yield()
        if not isinstance(space, ZfitSpace):
            try:
                space = pdf.space.with_binning(space)
            except Exception as error:
                msg = f"Could not create space {space} from pdf {pdf} with binning {space}"
                raise ValueError(msg) from error
        if label is None:
            label = f"Binned_{pdf.name}"
        super().__init__(
            obs=space,
            extended=extended,
            norm=norm,
            models=pdf,
            params={},
            name=name,
            label=label,
        )
        self.pdfs = self.models

    # def _get_params(self, floating: bool | None = True, is_yield: bool | None = None,
    #                 extract_independent: bool | None = True) -> set[ZfitParameter]:
    #     params = super()._get_params(floating=floating, is_yield=is_yield, extract_independent=extract_independent)
    #     daughter_params = self.pdfs[0].get_params(floating=floating, is_yield=is_yield,
    #                                               extract_independent=extract_independent)
    #     return daughter_params | params

    @z.function
    def _rel_counts(self, x, norm):
        del x  # not used, we just return the full histogram
        pdf = self.pdfs[0]
        edges = [znp.array(edge) for edge in self.axes.edges]
        edges_flat = [znp.reshape(edge, [-1]) for edge in edges]
        lowers = [edge[:-1] for edge in edges_flat]
        uppers = [edge[1:] for edge in edges_flat]
        lowers_meshed = znp.meshgrid(*lowers, indexing="ij")
        uppers_meshed = znp.meshgrid(*uppers, indexing="ij")
        shape = tf.shape(lowers_meshed[0])
        lowers_meshed_flat = [znp.reshape(lower_mesh, [-1]) for lower_mesh in lowers_meshed]
        uppers_meshed_flat = [znp.reshape(upper_mesh, [-1]) for upper_mesh in uppers_meshed]
        lower_flat = znp.stack(lowers_meshed_flat, axis=-1)
        upper_flat = znp.stack(uppers_meshed_flat, axis=-1)
        options = {"type": "bins"}

        @z.function
        def integrate_one(limits):
            low, up = tf.unstack(limits)
            limits_space = zfit.Space(obs=self.obs, limits=[low, up])
            return pdf.integrate(limits_space, norm=False, options=options)

        limits = znp.stack([lower_flat, upper_flat], axis=1)
        from zfit import run

        try:
            if run.get_graph_mode() is False:
                msg = "Just stearing the eager execution"
                raise MapNotVectorized(msg)
            values = tf.vectorized_map(integrate_one, limits)[:, 0]
        except (ValueError, MapNotVectorized):
            values = znp.asarray(tuple(map(integrate_one, limits)))
        values = znp.reshape(values, shape)
        if norm:
            values /= pdf.normalization(norm)
        return values

    @z.function(wraps="model_binned")
    def _counts(self, x, norm):
        del x  # not used, we just return the full histogram
        pdf = self.pdfs[0]
        edges = [znp.array(edge) for edge in self.axes.edges]
        edges_flat = [znp.reshape(edge, [-1]) for edge in edges]
        lowers = [edge[:-1] for edge in edges_flat]
        uppers = [edge[1:] for edge in edges_flat]
        lowers_meshed = znp.meshgrid(*lowers, indexing="ij")
        uppers_meshed = znp.meshgrid(*uppers, indexing="ij")
        shape = tf.shape(lowers_meshed[0])
        lowers_meshed_flat = [znp.reshape(lower_mesh, [-1]) for lower_mesh in lowers_meshed]
        uppers_meshed_flat = [znp.reshape(upper_mesh, [-1]) for upper_mesh in uppers_meshed]
        lower_flat = znp.stack(lowers_meshed_flat, axis=-1)
        upper_flat = znp.stack(uppers_meshed_flat, axis=-1)
        options = {"type": "bins"}

        if pdf.is_extended:

            @z.function
            def integrate_one(limits):
                low, up = tf.unstack(limits)
                limits_space = zfit.Space(obs=self.obs, limits=[low, up])
                return pdf.ext_integrate(limits_space, norm=False, options=options)

            missing_yield = False
        else:

            @z.function
            def integrate_one(limits):
                low, up = tf.unstack(limits)
                limits_space = zfit.Space(obs=self.obs, limits=[low, up])
                return pdf.integrate(limits_space, norm=False, options=options)

            missing_yield = True

        limits = znp.stack([lower_flat, upper_flat], axis=1)
        from zfit import run

        try:
            if run.get_graph_mode() is False:
                msg = "Just stearing the eager execution"
                raise MapNotVectorized(msg)
            values = tf.vectorized_map(integrate_one, limits)[:, 0]
        except (ValueError, MapNotVectorized):
            values = znp.asarray(tuple(map(integrate_one, limits)))
        values = znp.reshape(values, shape)
        if missing_yield:
            values *= self.get_yield()
        if norm:
            values /= pdf.normalization(norm)
        return values

    def __str__(self):
        return f"<BinnedFromUnbinnedPDF model={self.pdfs[0]} binning={self.space.binning}>"



#  Copyright (c) 2022 zfit
import numpy as np
import tensorflow as tf
from scipy.special import wofz

from zfit import z
import zfit.z.numpy as znp


def test_z_numpy_ndarray_is_tensorflow_tensor():
    """In tensorflow 2.4.1 tf.experimental.numpy.ndarray was a wrapper around tf.Tensor.

    Now this concept seems to have been scratched and tf.experimental.numpy.ndarray is just an alias for tf.Tensor. See
    the commit history of
    https://github.com/tensorflow/tensorflow/commits/master/tensorflow/python/ops/numpy_ops/np_arrays.py
    """
    assert znp.ndarray is tf.Tensor
    assert isinstance(znp.array(1), tf.Tensor)
    assert isinstance(znp.sum(znp.array(0)), tf.Tensor)


def test_faddeeva():
    """We only test with positive imaginary part as the Humlicek approximation of the Faddeeva function is becoming
    unstable in the negative imaginary part region.

    That is also due to the nature of the Faddeeva function in that region.
    """
    size = int(1e4)
    test_values1 = np.random.uniform(0, 0.1, size) + 1.0j * np.random.uniform(
        0, 0.1, size
    )
    test_values2 = np.random.uniform(0, 1, size) + 1.0j * np.random.uniform(0, 1, size)
    test_values3 = np.random.uniform(0, 10, size) + 1.0j * np.random.uniform(
        0, 10, size
    )
    test_values4 = np.random.uniform(0, 100, size) + 1.0j * np.random.uniform(
        0, 100, size
    )
    test_values5 = np.random.uniform(-0.1, 0, size) + 1.0j * np.random.uniform(
        0, 1, size
    )
    test_values6 = np.random.uniform(-1, 0, size) + 1.0j * np.random.uniform(0, 1, size)
    test_values7 = np.random.uniform(-10, 0, size) + 1.0j * np.random.uniform(
        0, 10, size
    )
    test_values8 = np.random.uniform(-100, 0, size) + 1.0j * np.random.uniform(
        0, 100, size
    )
    test_values = np.concatenate(
        [
            test_values1,
            test_values2,
            test_values3,
            test_values4,
            test_values5,
            test_values6,
            test_values7,
            test_values8,
        ]
    )
    np.testing.assert_allclose(
        znp.faddeeva_humlicek(test_values), wofz(test_values), rtol=1e-06
    )

    test_values_tensor = znp.asarray(test_values)
    with tf.GradientTape(persistent=True, watch_accessed_variables=False) as tape:
        tape.watch(test_values_tensor)
        result = znp.faddeeva_humlicek(test_values_tensor)
    gradientstf = tape.gradient(result, test_values_tensor)
    assert np.all(np.isfinite(gradientstf))

    # we test on a small sample of size 10 because the numerical gradient is slow to compute
    # we restrict our selves to real x and not go far away from 0 (-1, 1) because the Faddeeva function flattens out and gradients become zero.
    # we test the real and imaginary part separately because numdifftools can't differentiate with respect to complex numbers.
    x = np.random.uniform(-1, 1, size=10)
    f_real = lambda x: znp.real(znp.faddeeva_humlicek(x))
    f_imag = lambda x: znp.imag(znp.faddeeva_humlicek(x))

    num_gradients_real = z.math.numdifftools.Gradient(f_real)(x)
    num_gradients_imag = z.math.numdifftools.Gradient(f_imag)(x)

    x = znp.asarray(x)
    with tf.GradientTape(persistent=True, watch_accessed_variables=False) as tape:
        tape.watch(x)
        result_real = f_real(x)
        result_imag = f_imag(x)
    tf_gradients_real = tape.gradient(result_real, x)
    tf_gradients_imag = tape.gradient(result_imag, x)

    np.testing.assert_allclose(np.diag(num_gradients_real), tf_gradients_real)
    np.testing.assert_allclose(np.diag(num_gradients_imag), tf_gradients_imag)



#  Copyright (c) 2024 zfit

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    pass  # ruff: noqa: F401

from collections.abc import Mapping

import tensorflow as tf

import zfit.z.numpy as znp

from ..core.interfaces import ZfitBinnedPDF
from ..core.space import supports
from ..util import ztyping
from ..util.exception import SpecificFunctionNotImplemented
from .binned_functor import BaseBinnedFunctorPDF


class BinwiseScaleModifier(BaseBinnedFunctorPDF):
    def __init__(
        self,
        pdf: ZfitBinnedPDF,
        modifiers: bool | Mapping[str, ztyping.ParamTypeInput] | None = None,
        extended: ztyping.ExtendedInputType = None,
        norm: ztyping.NormInputType = None,
        name: str | None = "BinnedTemplatePDF",
        label: str | None = None,
    ) -> None:
        """Modifier that scales each bin separately of the *pdf*.

        Binwise modification can be used to account for uncorrelated or correlated uncertainties.

        Args:
            pdf: Binned pdf to be modified.
            modifiers: Modifiers for each bin.
            extended: |@doc:pdf.init.extended| The overall yield of the PDF.
               If this is parameter-like, it will be used as the yield,
               the expected number of events, and the PDF will be extended.
               An extended PDF has additional functionality, such as the
               ``ext_*`` methods and the ``counts`` (for binned PDFs). |@docend:pdf.init.extended|
            norm: |@doc:pdf.init.norm| Normalization of the PDF.
               By default, this is the same as the default space of the PDF. |@docend:pdf.init.norm|
            name: |@doc:pdf.init.label| Human-readable name
               or label of
               the PDF for a better description, to be used with plots etc.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.label|
            label: |@doc:pdf.init.label| Human-readable name
               or label of
               the PDF for a better description, to be used with plots etc.
               Has no programmatical functional purpose as identification. |@docend:pdf.init.label|
        """
        obs = pdf.space
        if not isinstance(pdf, ZfitBinnedPDF):
            msg = "pdf must be a BinnedPDF"
            raise TypeError(msg)
        if extended is None:
            extended = pdf.is_extended
        if modifiers is None:
            modifiers = True
        if modifiers is True:
            import zfit

            modifiers = {
                f"sysshape_{i}": zfit.Parameter(f"auto_sysshape_{self}_{i}", 1.0)
                for i in range(pdf.counts(obs).shape.num_elements())
            }
        if not isinstance(modifiers, dict):
            msg = "modifiers must be a dict-like object or True or None"
            raise TypeError(msg)
        params = modifiers.copy()
        self._binwise_modifiers = modifiers
        if extended is True:
            self._automatically_extended = True
            if modifiers:
                import zfit

                def sumfunc(params):
                    del params  # unused
                    values = self.counts()
                    return znp.sum(values)

                from zfit.core.parameter import get_auto_number

                params_sumfunc = modifiers.copy()
                dep_params = {}
                for p in pdf.get_params():
                    while True:
                        i = get_auto_number()
                        name_tmp = f"DUMMPY_PARAM_{i}"
                        if name_tmp not in params_sumfunc:
                            dep_params[name_tmp] = p
                            break

                params_sumfunc.update(dep_params)
                extended = zfit.ComposedParameter(
                    f"AUTO_binwise_modifier_{get_auto_number()}",
                    sumfunc,
                    params=params_sumfunc,
                )

            else:
                extended = pdf.get_yield()
        elif extended is not False:
            self._automatically_extended = False
        super().__init__(obs=obs, name=name, params=params, models=pdf, extended=extended, norm=norm, label=label)

    @supports(norm=True)
    def _counts(self, x, norm=None):
        if not self._automatically_extended:
            raise SpecificFunctionNotImplemented
        return self._counts_with_modifiers(x, norm)

    def _counts_with_modifiers(self, x, norm):
        values = self.pdfs[0].counts(x, norm=norm)
        if modifiers := list(self._binwise_modifiers.values()):
            sysshape_flat = tf.stack(modifiers)
            modifiers = znp.reshape(sysshape_flat, values.shape)
            values = values * modifiers
        return values

    @supports(norm="space")
    def _rel_counts(self, x, norm=None):
        values = self._counts_with_modifiers(x, norm)
        return values / znp.sum(values)



#  Copyright (c) 2024 zfit
import numpy as np
import pytest
import zfit
from zfit import z
import zfit.z.numpy as znp
from scipy.stats import gamma as gamma_scipy
import tensorflow as tf


gamma_true = 5.0
beta_true = 0.8
mu_true = 1.0

def create_gamma(gamma, beta, mu, limits):
    obs = zfit.Space("obs1", limits)
    gamma = zfit.pdf.Gamma(gamma=gamma_true, beta=beta, mu=mu, obs=obs)
    return gamma, obs


def test_gamma_pdf():
    gamma, _ = create_gamma(gamma=gamma_true, beta=beta_true, mu=mu_true, limits=(1, 10))
    assert pytest.approx(
        gamma_scipy.pdf(4, a=gamma_true, scale=beta_true, loc=mu_true), rel=1e-5
    ) == gamma.pdf(4, norm=False)
    test_values = znp.linspace(1, 10, 10_000)
    test_values_ones = np.ones_like(test_values)
    np.testing.assert_allclose(
        gamma.pdf(test_values, norm=False),
        gamma_scipy.pdf(test_values, a=gamma_true, scale=beta_true, loc=mu_true),
        rtol=1e-5,
    )
    np.testing.assert_array_less(gamma.pdf(test_values, norm=False), test_values_ones * gamma.pdf(4.2, norm=False))

    sample = gamma.sample(10_00)
    assert np.all(np.isfinite(sample.value())), "Some samples from the gamma PDF are NaN or infinite"
    assert sample.n_events == 1000
    ones_like = np.ones_like(sample.value())
    np.testing.assert_array_less(1 * ones_like, sample.value())
    np.testing.assert_array_less(sample.value(), ones_like * 10)


def test_gamma_integral():
    qgauss, obs = create_gamma(gamma=gamma_true, beta=beta_true, mu=mu_true, limits=(1, 10))
    full_interval_analytic = qgauss.analytic_integrate(obs, norm=False)
    full_interval_numeric = qgauss.numeric_integrate(obs, norm=False)
    scipy_full_inttegral = gamma_scipy.cdf(10, a=gamma_true, scale=beta_true, loc=mu_true) - gamma_scipy.cdf(
        1, a=gamma_true, scale=beta_true, loc=mu_true
    )
    assert pytest.approx(scipy_full_inttegral, 1e-6) == full_interval_analytic
    assert pytest.approx(scipy_full_inttegral, 1e-6) == full_interval_numeric

    analytic_integral = qgauss.analytic_integrate(limits=(3, 6), norm=False)
    numeric_integral = qgauss.numeric_integrate(limits=(3, 6), norm=False)
    scipy_integral = gamma_scipy.cdf(6, a=gamma_true, scale=beta_true, loc=mu_true) - gamma_scipy.cdf(
        3, a=gamma_true, scale=beta_true, loc=mu_true
    )
    assert pytest.approx(numeric_integral, 1e-5) == analytic_integral
    assert pytest.approx(scipy_integral, 1e-5) == analytic_integral
